#!/usr/bin/env python
"""Command line wrappers"""
import os
import sys
from collections import deque

import click
from colorama import Fore, Style, init
from loguru import logger

import genomepy

init(autoreset=True)
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(genomepy.__version__, "-v", "--version")
def cli():
    pass  # noqa


@click.command(short_help="show 1st lines of each annotation")
@click.argument("name")
@click.option("-p", "--provider", help="only search this provider")
@click.option("-n", "--lines", help="number of lines to print", default=2)
def annotation(name, provider=None, lines=None):
    """
    Quickly inspect the metadata of each GTF annotation available
    for the given genome.

    For UCSC, up to 4 gene annotation styles are available:
    "ncbiRefSeq", "refGene", "ensGene", "knownGene" (respectively).

    For NCBI, the chromosome names are not yet sanitized.
    """
    genomepy.head_annotations(name, provider, n=int(lines))


@click.command("clean", short_help="remove provider data")
def clean():
    """
    Remove cached data on providers (e.g. available genomes).
    """
    genomepy.clean()


@click.command("config", short_help="manage configuration")
@click.argument("command")
def config(command):
    """
    Manage configuration

    genomepy config file        return config filepath

    genomepy config show        return config content

    genomepy config generate    create new config file
    """
    genomepy.manage_config(command)


@click.command("genomes", short_help="list available genomes")
@click.option("-p", "--provider", help="provider")
@click.option("-s", "--size", is_flag=True, help="show absolute genome size")
def genomes(provider=None, size=False):
    """
    List all available genomes.

    Returns the metadata of each found genome, including the availability of a gene annotation.
    For UCSC, up to 4 gene annotation styles are available:
    "ncbiRefSeq", "refGene", "ensGene", "knownGene" (respectively).
    """
    provider_init = True
    for row in genomepy.list_available_genomes(provider, size):
        if provider_init:
            provider_init = False
            terminal_header(size)
            if str(provider).lower() in ["none", "ucsc"]:
                terminal_subheader(size)
        terminal_formatting(row)


# extended options for genomepy install
INSTALL_OPTIONS = {
    "provider": {
        "short": "p",
        "long": "provider",
        "help": "download from this provider",
        "default": None,
    },
    "genomes_dir": {
        "short": "g",
        "long": "genomes_dir",
        "help": "create output directory here",
        "default": None,
    },
    "localname": {
        "short": "l",
        "long": "localname",
        "help": "custom name",
        "default": None,
    },
    "mask": {
        "short": "m",
        "long": "mask",
        "help": "DNA masking: hard/soft/none (default: soft)",
        "default": "soft",
    },
    "keep_alt": {
        "short": "k",
        "long": "keep-alt",
        "help": "keep alternative regions",
        "flag_value": True,
    },
    "regex": {
        "short": "r",
        "long": "regex",
        "help": "regex to filter sequences",
        "default": None,
    },
    "invert_match": {
        "short": "n",
        "long": "no-match",
        "help": "select sequences that *don't* match regex",
        "flag_value": True,
    },
    "bgzip": {
        "short": "b",
        "long": "bgzip",
        "help": "bgzip genome",
        "flag_value": True,
    },
    "threads": {
        "short": "t",
        "long": "threads",
        "help": "build index using multithreading",
        "default": min(os.cpu_count(), 8),
    },
    "force": {
        "short": "f",
        "long": "force",
        "help": "overwrite existing files",
        "flag_value": True,
    },
    "text_line1": {
        "long": "Annotation options:",
        "help": "",
        "flag_value": True,
        "text_line": True,
    },
    "annotation": {
        "short": "a",
        "long": "annotation",
        "help": "download annotation",
        "flag_value": True,
    },
    "only_annotation": {
        "short": "o",
        "long": "only_annotation",
        "help": "only download annotation (sets -a)",
        "flag_value": True,
    },
    "skip_matching": {
        "short": "sm",
        "long": "skip_matching",
        "help": "skip matching contigs between the gene annotation and the genome (sets -a)",
        "flag_value": True,
    },
    "skip_filter": {
        "short": "sf",
        "long": "skip_filter",
        "help": "skip filtering out contigs in the gene annotation missing from the genome (sets -a)",
        "flag_value": True,
    },
    "text_line2": {
        "long": "Provider specific options:",
        "help": "",
        "flag_value": True,
        "text_line": True,
    },
}


def get_install_options():
    """
    Combine general and provider specific options.

    Add the provider name in front of the options to prevent overlap.
    """
    if len(sys.argv) > 1 and sys.argv[1] == "install":
        install_options = INSTALL_OPTIONS

        # extend install options with provider specific options
        for provider in genomepy.list_providers():
            p_dict = eval(
                "genomepy.providers."
                + provider.capitalize()
                + "Provider._cli_install_options"
            )
            for option in p_dict.keys():
                p_dict[option]["long"] = provider + "-" + p_dict[option]["long"]
            install_options.update(p_dict)

        return install_options
    return {}


def custom_options(options):
    """Dynamically add options to a click.command (based on a dict with options)."""

    def decorator(f):
        for opt_name, opt_params in options.items():
            param_decls = deque(["--" + opt_params["long"], opt_name])
            if "short" in opt_params.keys():
                param_decls.appendleft("-" + opt_params["short"])

            attrs = dict(help=opt_params["help"])
            if "type" in opt_params.keys():
                attrs["type"] = opt_params["type"]
            if "default" in opt_params.keys():
                attrs["default"] = opt_params["default"]
            if "flag_value" in opt_params.keys():
                attrs["flag_value"] = opt_params["flag_value"]

            # can be used to add paragraphs to the --help menu
            if "text_line" in opt_params.keys():
                param_decls = deque(["\n" + opt_params["long"], opt_name])

            click.option(*param_decls, **attrs)(f)
        return f

    return decorator


@custom_options(get_install_options())
@click.argument("name")
@cli.command(short_help="install a genome & run active plugins")
def install(
    name,
    provider,
    genomes_dir,
    localname,
    mask,
    keep_alt,
    regex,
    invert_match,
    bgzip,
    annotation,
    only_annotation,
    skip_matching,
    skip_filter,
    threads,
    force,
    **kwargs,
):
    """
    Install a genome & run active plugins.

    NAME (and more) can be obtained from genomepy search.
    """
    genomepy.install_genome(
        name,
        provider=provider,
        genomes_dir=genomes_dir,
        localname=localname,
        mask=mask,
        keep_alt=keep_alt,
        regex=regex,
        invert_match=invert_match,
        bgzip=bgzip,
        annotation=annotation,
        only_annotation=only_annotation,
        skip_matching=skip_matching,
        skip_filter=skip_filter,
        threads=threads,
        force=force,
        **kwargs,
    )


@click.command("plugin", short_help="manage plugins")
@click.argument("command")
@click.argument("name", nargs=-1)
def plugin(command, name):
    """
    Enable or disable plugins.

    genomepy plugin list                 show plugins and status

    genomepy plugin enable  [NAME(S)]    enable plugins

    genomepy plugin disable [NAME(S)]    disable plugins
    """
    genomepy.manage_plugins(command, name)


@click.command("providers", short_help="list available providers")
def providers():
    """List all available providers."""
    for p in genomepy.list_providers():
        print(p)


# names and sizes for the search output columns (when connected to a terminal)
SEARCH_FORMAT = {
    "name": "<20",
    "provider": "<8",  # fixed width
    "accession": "<16",  # fixed width
    "tax_id": ">7",  # fixed width
    "annotation": "^10",  # fixed width
    "species": "<40",
    "genome_size": "<13",
    "other_info": "<40",
}
FULL_SEARCH_STRING = " ".join([f"{{:{size}}}" for size in SEARCH_FORMAT.values()])
SEARCH_STRING = " ".join(
    [f"{{:{size}}}" for size in SEARCH_FORMAT.values() if size != "<13"]
)
if sys.stdout.isatty():

    def bool_to_unicode(boolean: bool) -> str:
        """Converts True to a checkmark and False to a cross-mark."""
        return "\u2713" if boolean else "\u2717"

    def color_unicode(string):
        """Color checkmark green and cross-mark red."""
        sting = string.replace("\u2713", Fore.GREEN + "\u2713" + Fore.RESET)
        sting = sting.replace("\u2717", Fore.RED + "\u2717" + Fore.RESET)
        return sting

    def terminal_formatting(row: list):
        """
        In case we print to a terminal, the output is aligned.
        Otherwise, (file, pipe) we use tab-separated columns.
        """
        if isinstance(row[4], list):
            # UCSC annotations
            row[4] = " ".join([bool_to_unicode(b) for b in row[4]])
        else:
            row[4] = bool_to_unicode(row[4])
        for n, ele in enumerate(row):
            if ele is None:
                row[n] = "na"
        if len(row) == 8:
            # genome_size
            row[6] = f"{int(row[6]):,}"
            row = FULL_SEARCH_STRING.format(*row)
        else:
            row = SEARCH_STRING.format(*row)
        print(color_unicode(row))

    def terminal_header(size):
        """Header for search output."""
        if size:
            print(Style.BRIGHT + FULL_SEARCH_STRING.format(*SEARCH_FORMAT))
        else:
            fmt = [k for k in SEARCH_FORMAT if k != "genome_size"]
            print(Style.BRIGHT + SEARCH_STRING.format(*fmt))

    def terminal_subheader(size):
        """Subheader for search output."""
        # annotations: ncbiRefSeq, refGene, ensGene & knownGene
        subheader = ["", "", "", "", "n r e k", "<- UCSC options (see help)", "", ""]
        if size:
            print(FULL_SEARCH_STRING.format(*subheader))
        else:
            print(SEARCH_STRING.format(*subheader[:-1]))

else:

    def terminal_formatting(row: list):
        """
        In case we print to a terminal, the output is aligned.
        Otherwise, (file, pipe) we use tab-separated columns.
        """
        if isinstance(row[4], list):
            row[4] = str(row[4])
        print("\t".join([str(element) for element in row]))

    def terminal_header(size):
        """Header for search output."""
        if size:
            print("\t".join(SEARCH_FORMAT))
        else:
            fmt = [k for k in SEARCH_FORMAT if k != "genome_size"]
            print("\t".join(fmt))

    def terminal_subheader(_):
        pass


@click.command(short_help="search for genomes")
@click.argument("term", nargs=-1)
@click.option("-p", "--provider", help="only search this provider")
@click.option("-s", "--size", is_flag=True, help="show absolute genome size")
def search(term, provider=None, size=False):
    """
    Search for genomes that contain TERM in their name, description,
    accession (must start with GCA_ or GCF_) or taxonomy (start).

    Search is case-insensitive, name/description search accepts multiple terms and regex.

    Returns the metadata of each found genome, including the availability of a gene annotation.
    For UCSC, up to 4 gene annotation styles are available:
    "ncbiRefSeq", "refGene", "ensGene", "knownGene" (respectively).
    Each with different naming schemes.
    """
    term = " ".join(term)
    no_genomes = True
    for row in genomepy.search(term, provider, size):
        if no_genomes:
            no_genomes = False
            terminal_header(size)
            if str(provider).lower() in ["none", "ucsc"]:
                terminal_subheader(size)
        terminal_formatting(row)

    if sys.stdout.isatty():
        if no_genomes:
            logger.warning("No genomes found!")
        else:
            print(Fore.GREEN + " ^")
            print(Fore.GREEN + " Use name for " + Fore.CYAN + "genomepy install")


cli.add_command(annotation)
cli.add_command(clean)
cli.add_command(config)
cli.add_command(genomes)
cli.add_command(install)
cli.add_command(plugin)
cli.add_command(providers)
cli.add_command(search)

if __name__ == "__main__":
    cli()



import os
from shutil import copyfile

from appdirs import user_config_dir
from norns import config as cfg

__all__ = ["config", "manage_config"]

config = cfg("genomepy", default="config/default.yaml")


def generate_config():
    config_dir = user_config_dir("genomepy")
    new_config = os.path.join(config_dir, "genomepy.yaml")

    # existing config must be removed before norns picks up the default again
    os.makedirs(config_dir, exist_ok=True)
    if os.path.exists(new_config):
        os.remove(new_config)

    default_config = cfg("genomepy", default="config/default.yaml").config_file
    copyfile(default_config, new_config)
    config.config_file = new_config
    print(f"Created config file {new_config}")


def manage_config(command):
    """
    Manage the genomepy configuration

    Parameters
    ----------
    command : str
        command to perform. Options:

        file
            return config filepath
        show
            return config content
        generate
            create new config file
    """
    if command in ["file", "path"]:
        print(config.config_file)

    elif command in ["show", "list"]:
        with open(config.config_file) as f:
            print(f.read())

    elif command == "generate":
        generate_config()

    else:
        raise ValueError(
            f"Invalid config command: '{command}'. "
            "Options: 'file', 'show' or 'generate'."
        )



import os

from genomepy.plugins import Plugin
from genomepy.utils import cmd_ok, mkdir_p, rm_rf, run_index_cmd


class Bowtie2Plugin(Plugin):
    def after_genome_download(self, genome, threads=1, force=False):
        if not cmd_ok("bowtie2-build"):
            return

        # Create index dir
        index_dir = genome.plugin["bowtie2"]["index_dir"]
        index_name = genome.plugin["bowtie2"]["index_name"]
        if force:
            # Start from scratch
            rm_rf(index_dir)
        mkdir_p(index_dir)

        if not any(fname.endswith(".bt2") for fname in os.listdir(index_dir)):
            # Create index
            cmd = f"bowtie2-build --threads {threads} {genome.filename} {index_name}"
            run_index_cmd("bowtie2", cmd)

    def get_properties(self, genome):
        props = {
            "index_dir": os.path.join(
                os.path.dirname(genome.filename), "index", "bowtie2"
            ),
            "index_name": os.path.join(
                os.path.dirname(genome.filename), "index", "bowtie2", genome.name
            ),
        }
        return props



import os
from shutil import rmtree
from time import time

from appdirs import user_cache_dir
from diskcache import Cache
from filelock import FileLock

from genomepy.__about__ import __version__
from genomepy.config import config
from genomepy.utils import rm_rf


def _cast(var):
    if str(var).lower() == "none":
        return None
    try:
        return float(var)
    except ValueError:
        return var


# Set short/long cache expiration times (in seconds)
cache_exp_other = _cast(config.get("cache_exp_other", 3.6e3))
cache_exp_genomes = _cast(config.get("cache_exp_genomes", 6.048e5))

genomepy_cache_dir = os.path.join(user_cache_dir("genomepy"), __version__)
os.makedirs(genomepy_cache_dir, exist_ok=True)

# create a lock, so only one tread can access the cache at once
lock_file = os.path.join(genomepy_cache_dir, "cache.lock")
if os.path.exists(lock_file) and time() - os.stat(lock_file).st_mtime > 60:
    # remove abandoned lock
    rm_rf(lock_file)
lock = FileLock(lock_file)

with lock:
    # Store the output of slow commands (marked with @disk_cache.memoize) for fast reuse
    # DiskCache uses the LRU (least-recently-stored) eviction policy by default
    disk_cache = Cache(directory=genomepy_cache_dir)


@lock
def clean():
    """Remove cached data on providers."""
    rmtree(genomepy_cache_dir, ignore_errors=True)
    os.makedirs(genomepy_cache_dir, exist_ok=True)
    print("All clean!")



import re

import requests
from loguru import logger

from genomepy.caching import cache_exp_genomes, cache_exp_other, disk_cache, lock
from genomepy.exceptions import GenomeDownloadError
from genomepy.online import check_url, retry
from genomepy.providers.base import BaseProvider
from genomepy.utils import safe


class EnsemblProvider(BaseProvider):
    """
    Ensembl genome provider.

    Will search both ensembl.org as well as ensemblgenomes.org.
    The bacteria division is not yet supported.
    """

    name = "Ensembl"
    accession_fields = ["assembly_accession"]
    taxid_fields = ["taxonomy_id"]
    description_fields = [
        "name",
        "scientific_name",
        "url_name",
        "display_name",
    ]
    _cli_install_options = {
        "toplevel": {
            "long": "toplevel",
            "help": "always download toplevel-genome",
            "flag_value": True,
        },
        "version": {
            "long": "version",
            "help": "select release version",
            "type": int,
            "default": None,
        },
    }
    _url = "https://rest.ensembl.org/"

    def __init__(self):
        self._provider_status()
        # Populate on init, so that methods can be cached
        self.genomes = get_genomes(self._url)

    @staticmethod
    def ping():
        """Can the provider be reached?"""
        api_online = bool(check_url("https://rest.ensembl.org/info/ping?"))
        vertebrate_url_online = bool(check_url("http://ftp.ensembl.org"))
        other_url_online = bool(check_url("http://ftp.ensemblgenomes.org"))
        return api_online and vertebrate_url_online and other_url_online

    def _genome_info_tuple(self, name, size=False):
        """tuple with assembly metadata"""
        accession = self.assembly_accession(name)
        taxid = self.genome_taxid(name)
        annotations = bool(self.annotation_links(name))
        species = self.genomes[name].get("scientific_name")
        other = self.genomes[name].get("genebuild")
        if size:
            length = self.genomes[name]["base_count"]
            return name, accession, taxid, annotations, species, length, other
        return name, accession, taxid, annotations, species, other

    @staticmethod
    @lock
    @disk_cache.memoize(expire=cache_exp_other, tag="get_version-ensembl")
    def get_version(vertebrates=False, set_version=None, url=_url):
        """Retrieve current version from Ensembl FTP."""
        if set_version:
            return str(set_version)
        ext = "/info/data/?" if vertebrates else "/info/eg_version?"
        ret = retry(request_json, 3, url, ext)
        releases = ret["releases"] if vertebrates else [ret["version"]]
        return str(max(releases))

    def get_genome_download_link(self, name, mask="soft", **kwargs):
        """
        Return http link to the genome sequence

        Parameters
        ----------
        name : str
            Genome name. Current implementation will fail if exact
            name is not found.

        mask : str , optional
            Masking level. Options: soft, hard or none. Default is soft.

        Returns
        ------
        str with the http download link.
        """
        genome = self.genomes[safe(name)]
        division, is_vertebrate = get_division(genome)

        # base directory of the genome
        ftp = "http://ftp.ensemblgenomes.org"
        if is_vertebrate:
            ftp = "http://ftp.ensembl.org"
        version = self.get_version(is_vertebrate, kwargs.get("version"))
        div_path = "" if is_vertebrate else f"/{division}"
        lwr_name = genome["url_name"].lower()

        ftp_directory = f"{ftp}/pub/release-{version}{div_path}/fasta/{lwr_name}/dna"
        # some entries don't use url_name in their url... -,-
        # examples:
        #   - EnsemblVertebrates: mus_musculus_nzohlltj
        #   - EnsemblMetazoa: caenorhabditis_elegans
        if not check_url(ftp_directory, 2):
            lwr_name = genome["name"]
            ftp_directory = (
                f"{ftp}/pub/release-{version}{div_path}/fasta/{lwr_name}/dna"
            )

        # this assembly has its own directory
        if name == "GRCh37":
            ftp_directory = genome["genome"].format(version)

        # specific fasta file
        cap_name = lwr_name.capitalize()
        asm_name = re.sub(r"\.p\d+$", "", safe(genome["assembly_name"]))
        mask_lvl = {"soft": "_sm", "hard": "_rm", "none": ""}[mask]
        asm_lvl = "toplevel" if kwargs.get("toplevel") else "primary_assembly"
        version_tag = "" if int(version) > 30 else f".{version}"

        ftp_file = f"{cap_name}.{asm_name}{version_tag}.dna{mask_lvl}.{asm_lvl}.fa.gz"

        # combine
        link = f"{ftp_directory}/{ftp_file}"
        if check_url(link, 2):
            return link

        # primary assemblies do not always exist
        if asm_lvl == "primary_assembly":
            link = link.replace("primary_assembly", "toplevel")
            if check_url(link, 2):
                return link

        raise GenomeDownloadError(
            f"Could not download genome {name} from {self.name}.\n"
            "URL is broken. Select another genome or provider.\n"
            f"Broken URL: {link}"
        )

    def get_annotation_download_links(self, name, **kwargs):
        """
        Retrieve functioning gene annotation download link(s).

        Parameters
        ----------
        name : str
            genome name
        **kwargs: dict, optional:
            version : Ensembl version to use. By default the latest version is used

        Returns
        -------
        list
            http link(s)
        """
        genome = self.genomes[safe(name)]
        division, is_vertebrate = get_division(genome)

        # base directory of the genome
        ftp = "http://ftp.ensemblgenomes.org"
        if is_vertebrate:
            ftp = "http://ftp.ensembl.org"
        version = self.get_version(is_vertebrate, kwargs.get("version"))
        div_path = "" if is_vertebrate else f"/{division}"
        lwr_name = genome["url_name"].lower()

        ftp_directory = f"{ftp}/pub/release-{version}{div_path}/gtf/{lwr_name}"
        # some entries don't use url_name in their url... -,-
        # examples:
        #   - EnsemblVertebrates: mus_musculus_nzohlltj
        #   - EnsemblMetazoa: caenorhabditis_elegans
        if not check_url(ftp_directory, 2):
            lwr_name = genome["name"]
            ftp_directory = f"{ftp}/pub/release-{version}{div_path}/gtf/{lwr_name}"

        # specific gtf file
        cap_name = lwr_name.capitalize()
        asm_name = re.sub(r"\.p\d+$", "", safe(genome["assembly_name"]))

        ftp_file = f"{cap_name}.{asm_name}.{version}.gtf.gz"

        # combine
        link = f"{ftp_directory}/{ftp_file}"
        if name == "GRCh37":
            link = genome["annotation"].format(version)
        return [link] if check_url(link, max_tries=2) else []


def get_division(genome: dict):
    """Retrieve the division of a genome."""
    division = genome["division"].lower().replace("ensembl", "")
    if division == "bacteria":
        raise NotImplementedError("Bacteria from Ensembl not supported.")

    is_vertebrate = bool(division == "vertebrates")
    return division, is_vertebrate


def request_json(rest_url, ext):
    """Make a REST request and return as json."""
    if rest_url.endswith("/") and ext.startswith("/"):
        ext = ext[1:]

    r = requests.get(rest_url + ext, headers={"Content-Type": "application/json"})

    if not r.ok:
        r.raise_for_status()

    return r.json()


def add_grch37(genomes):
    genomes["GRCh37"] = {
        "genome": (
            "http://ftp.ensembl.org/pub/grch37/release-{}/fasta/homo_sapiens/dna"
        ),
        "annotation": (
            "http://ftp.ensembl.org/pub/grch37/release-{}/gtf/"
            "homo_sapiens/Homo_sapiens.GRCh37.87.gtf.gz"
        ),
        "assembly_accession": "GCA_000001405.14",
        "taxonomy_id": 9606,
        "name": "human",
        "scientific_name": "Homo sapiens",
        "url_name": "Homo_sapiens",
        "assembly_name": "GRCh37",
        "division": "vertebrates",
        "base_count": "3137144693",
        "display_name": "",
        "genebuild": "",
    }
    return genomes


@lock
@disk_cache.memoize(expire=cache_exp_genomes, tag="get_genomes-ensembl")
def get_genomes(rest_url):
    logger.info("Downloading assembly summaries from Ensembl")

    genomes = {}
    divisions = retry(request_json, 3, rest_url, "info/divisions?")
    for division in divisions:
        if division == "EnsemblBacteria":
            continue
        division_genomes = retry(
            request_json, 3, rest_url, f"info/genomes/division/{division}?"
        )

        # filter summaries to these keys (to reduce the size of the cached data)
        summary_keys_to_keep = [
            "assembly_name",
            "assembly_accession",
            "taxonomy_id",
            "name",
            "scientific_name",
            "url_name",
            "display_name",
            "genebuild",
            "division",
            "base_count",
        ]
        for genome in division_genomes:
            name = safe(genome["assembly_name"])
            genomes[name] = {k: genome[k] for k in summary_keys_to_keep}

    genomes = add_grch37(genomes)
    return genomes



import os
import re
import subprocess as sp
import urllib.error
from tempfile import mkdtemp
from typing import Generator, Iterator, List

import mysql.connector
import pandas as pd
import requests
from loguru import logger

from genomepy.caching import cache_exp_genomes, disk_cache, lock
from genomepy.exceptions import GenomeDownloadError
from genomepy.files import update_readme
from genomepy.online import check_url, read_url
from genomepy.providers.base import BaseProvider
from genomepy.providers.ncbi import NcbiProvider, get_genome_size
from genomepy.utils import (
    check_ucsc_tools,
    get_genomes_dir,
    get_localname,
    lower,
    mkdir_p,
    rm_rf,
)

# order determines which annotation genomepy will attempt to install
# for more info, see http://genome.ucsc.edu/FAQ/FAQgenes.html
ANNOTATIONS = ["ncbiRefSeq", "refGene", "ensGene", "knownGene"]


class UcscProvider(BaseProvider):
    """
    UCSC genome provider.

    The UCSC API REST server is used to search and list genomes.
    The UCSC MySQL database is used to find metadata and annotations.
    """

    name = "UCSC"
    accession_fields = ["assembly_accession"]
    taxid_fields = ["taxId"]
    description_fields = ["description", "scientificName"]
    _cli_install_options = {
        "ucsc_annotation_type": {
            "long": "annotation",
            "help": "specify annotation to download: "
            f"{', '.join(ANNOTATIONS)} (case-insensitive)",
            "default": None,
        },
    }
    _url = "http://hgdownload.soe.ucsc.edu/goldenPath"

    def __init__(self):
        self._provider_status()
        # Populate on init, so that methods can be cached
        self.genomes = get_genomes("http://api.genome.ucsc.edu/list/ucscGenomes")

    @staticmethod
    def ping():
        """Can the provider be reached?"""
        url_online = bool(check_url("http://hgdownload.soe.ucsc.edu/goldenPath"))
        api_online = bool(check_url("http://api.genome.ucsc.edu/list/ucscGenomes"))
        return url_online and api_online

    def _search_accession(self, term: str) -> Iterator[str]:
        """
        UCSC does not always store assembly accessions.
        If no hits were found, search accession on NCBI (most genomes + stable accession IDs),
        then uses those search results for a UCSC text search.

        Parameters
        ----------
        term : str
            Assembly accession, `GCA_`/`GCF_`

        Yields
        ------
        genome names
        """
        # cut off prefix (GCA_/GCF_) and suffix (version numbers, e.g. '.3')
        term = term[4:].split(".")[0]
        hits = 0
        for name, metadata in self.genomes.items():
            if any([term in str(metadata[f]) for f in self.accession_fields]):
                hits += 1
                yield name

        # search NCBI only if we found no local hits
        if hits == 0:
            return self._search_accession_ncbi(term)

    def _search_accession_ncbi(self, term: str) -> Iterator[str]:
        """
        search NCBI (most genomes + stable accession IDs),
        then uses the NCBI accession search results for a UCSC text search.

        Parameters
        ----------
        term : str
            Assembly accession, `GCA_`/`GCF_`

        Yields
        ------
        genome names
        """
        # NCBI provides a consistent assembly accession. This can be used to
        # retrieve the species, and then search for that.
        p = NcbiProvider()
        ncbi_genomes = list(p._search_accession(term))

        # remove superstrings (keep GRCh38, not GRCh38.p1 to GRCh38.p13)
        unique_ncbi_genomes = []
        for i in ncbi_genomes:
            if sum([j in i for j in ncbi_genomes]) == 1:
                unique_ncbi_genomes.append(i)

        # add NCBI organism names to search terms
        organism_names = [
            p.genomes[name]["organism_name"] for name in unique_ncbi_genomes
        ]
        terms = list(set(unique_ncbi_genomes + organism_names))

        # search with NCBI results in the given provider
        for name, metadata in self.genomes.items():
            for term in terms:
                term = lower(term)
                if term in lower(name) or any(
                    [term in lower(metadata[f]) for f in self.description_fields]
                ):
                    yield name
                    break  # max one hit per genome

    def assembly_accession(self, name: str) -> str or None:
        """
        Return the assembly accession (`GCA_`/`GCF_`) for a genome.

        Some accession IDs can be retrieved from the UCSC MySQL hgFixed database.
        For others, the accession IDs can sometimes be scraped from the readme.html.
        If not, any linked NCBI assembly pages can also be scraped.

        Parameters
        ----------
        name: str
            genome name

        Returns
        ------
        str
            Assembly accession.
        """
        acc = self.genomes[name]["assembly_accession"]
        if acc:
            return acc

        acc = scrape_accession(self.genomes[name]["htmlPath"])
        if acc:
            return acc

    def annotation_links(self, name, **kwargs) -> List[str]:
        """
        Return a sorted list of available gene annotation types for a genome

        Parameters
        ----------
        name: str
            genome name

        Returns
        ------
        list
            Gene annotation types
        """
        available = self.get_annotation_download_links(name)
        annotations_found = [a for a in ANNOTATIONS if a in available]
        return annotations_found

    def _genome_info_tuple(self, name, size=False):
        """tuple with assembly metadata"""
        accession = self.assembly_accession(name)
        taxid = self.genome_taxid(name)
        annotations = [a in self.annotation_links(name) for a in ANNOTATIONS]
        species = self.genomes[name].get("scientificName")
        other = self.genomes[name].get("description")
        if size:
            length = get_genome_size(accession)
            return name, accession, taxid, annotations, species, length, other
        return name, accession, taxid, annotations, species, other

    def get_genome_download_link(self, name, mask="soft", **kwargs):
        """
        Return UCSC http link to genome sequence

        Parameters
        ----------
        name : str
            Genome name. Current implementation will fail if exact
            name is not found.

        mask : str , optional
            Masking level. Options: soft, hard or none. Default is soft.

        Returns
        ------
        str with the http/ftp download link.
        """
        ucsc_url = self._url + "/{0}/bigZips/chromFa.tar.gz"
        ucsc_url_masked = self._url + "/{0}/bigZips/chromFaMasked.tar.gz"
        alt_ucsc_url = self._url + "/{0}/bigZips/{0}.fa.gz"
        alt_ucsc_url_masked = self._url + "/{0}/bigZips/{0}.fa.masked.gz"

        # soft masked genomes. can be unmasked in _post _process_download
        urls = [ucsc_url, alt_ucsc_url]
        if mask == "hard":
            urls = [ucsc_url_masked, alt_ucsc_url_masked]

        for genome_url in urls:
            link = genome_url.format(name)

            if check_url(link, 2):
                return link

        raise GenomeDownloadError(
            f"Could not download genome {name} from {self.name}.\n"
            "URLs are broken. Select another genome or provider.\n"
            f"Broken URLs: {', '.join([url.format(name) for url in urls])}"
        )

    @staticmethod
    def _post_process_download(name, fname, out_dir, mask="soft"):  # noqa
        """
        Unmask a softmasked genome if required

        Parameters
        ----------
        name : str
            unused for the UCSC function

        fname : str
            file path to the genome fasta

        out_dir : str
            unused for the UCSC function

        mask : str , optional
            masking level: soft/hard/none, default=soft
        """
        if mask != "none":
            return

        logger.info("UCSC genomes are softmasked by default. Unmasking...")

        old_fname = os.path.join(
            os.path.dirname(fname), f"original_{os.path.basename(fname)}"
        )
        os.rename(fname, old_fname)
        with open(old_fname) as old, open(fname, "w") as new:
            for line in old:
                if line[0] == ">":
                    new.write(line)
                else:
                    new.write(line.upper())

    def get_annotation_download_links(self, name, **kwargs):
        """
        Return available gene annotation table(s) from the UCSC MySQL database.

        Available tables were retrieved on init.

        Parameters
        ----------
        name : str
            genome name

        Returns
        -------
        list
            annotation types
        """
        return self.genomes[name]["annotations"]

    def get_annotation_download_link(self, name: str, **kwargs) -> str:
        """
        Return an available annotation type.

        Parameters
        ----------
        name : str
            genome name
        **kwargs: dict, optional:
            ucsc_annotation_type : specific annotation type to download.

        Returns
        -------
        str
            http/ftp link

        Raises
        ------
        GenomeDownloadError
            if no functional link was found
        FileNotFoundError
            if the specified annotation type is unavailable
        """
        available = self.annotation_links(name)
        if not available:
            raise GenomeDownloadError(
                f"No gene annotations found for {name} on {self.name}.\n"
                "Check for typos or try\n"
                f"  genomepy search {name} -p {self.name}"
            )
        annot = available

        usr_annot = kwargs.get("ucsc_annotation_type")
        if usr_annot:
            # not all types are available for each genome
            annot = [a for a in available if a.lower() == usr_annot.lower()]
            if not annot:
                raise FileNotFoundError(
                    f"{usr_annot} is not available for {name}. "
                    f"Options: {', '.join(available)}.\n"
                )

        return annot[0]

    def download_annotation(self, name, genomes_dir=None, localname=None, **kwargs):
        """
        Download the UCSC genePred via their MySQL database, and convert to annotations.
        """
        name = self._check_name(name)
        annot = self.get_annotation_download_link(name, **kwargs)

        localname = get_localname(name, localname)
        genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)

        logger.info(f"Downloading the {annot} annotation from the UCSC MySQL database.")
        try:
            download_annotation(name, annot, genomes_dir, localname)
            logger.info("Annotation download successful")
        except Exception as e:
            raise GenomeDownloadError(
                f"An error occured while installing the gene annotation for {name} from {self.name}.\n"
                "If you think the annotation should be there, please file a bug report at: "
                "https://github.com/vanheeringen-lab/genomepy/issues\n\n"
                f"Error: {e.args[0]}"
            )

        # Add annotation URL to readme
        readme = os.path.join(genomes_dir, localname, "README.txt")
        update_readme(
            readme,
            updated_metadata={
                "annotation url": f"UCSC MySQL database: {name}, table: {annot}"
            },
        )

    def head_annotation(self, name, genomes_dir=None, n: int = 5, **kwargs):
        """
        Download the first n genes of each UCSC annotation type.

        The first line of the GTF is printed for review
        (of the gene_name field, for instance).

        Parameters
        ----------
        name : str
            genome name
        genomes_dir : str, optional
            genomes directory to install the annotation in.
        n : int, optional
            download the annotation for n genes.
        kwargs : dict , optional
            annotations : list
                specify which UCSC annotation types to download.
                Downloads all available if left blank.
        """
        name = self._check_name(name)
        genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)

        annotations = kwargs.get("annotations")
        if annotations is None:
            annotations = self.annotation_links(name)

        for annot in annotations:
            if annot not in ANNOTATIONS:
                raise ValueError(f"{annot} not in {ANNOTATIONS}")

            localname = f"{name}_head_{annot}"
            fpath = os.path.join(genomes_dir, localname, f"{localname}.annotation.gtf")
            download_annotation(name, annot, genomes_dir, localname, n=n)

            logger.info(f"{self.name} {annot}")
            with open(fpath) as f:
                for m, line in enumerate(f):
                    if line:
                        print(line.strip())
                        if m + 1 == n:
                            break


@lock
@disk_cache.memoize(expire=cache_exp_genomes, tag="get_genomes-ucsc")
def get_genomes(rest_url):
    logger.info("Downloading assembly summaries from UCSC")

    r = requests.get(rest_url, headers={"Content-Type": "application/json"})
    if not r.ok:
        r.raise_for_status()
    ucsc_json = r.json()
    genomes = ucsc_json["ucscGenomes"]

    # filter summaries to these keys (to reduce the size of the cached data)
    summary_keys_to_keep = [
        "taxId",
        "description",
        "scientificName",
        "sourceName",
        "htmlPath",
    ]
    for genome in genomes:
        genomes[genome] = {k: genomes[genome][k] for k in summary_keys_to_keep}
        genomes[genome]["assembly_accession"] = None
        genomes[genome]["annotations"] = []
    # add accession IDs (self.assembly_accession will try to fill in the blanks)
    genomes = add_accessions1(genomes)
    genomes = add_accessions2(genomes)
    genomes = add_annotation_links(genomes)

    return genomes


def add_accessions1(genomes: dict) -> dict:
    """
    Use the UCSC MySQL database to obtain accession IDs
    (of the most similar assemblies on other providers).

    For NCBI, RefSeq and GenBank accession IDs are stored directly in this table.

    Updates the the genome dict "assembly_accession" field for each genome found.
    """
    # MySQL query
    database = "hgFixed"
    command = (
        "SELECT source,destination,matchCount "
        "FROM asmEquivalent "
        "WHERE sourceAuthority='ucsc' "
        "AND destinationAuthority!='ensembl' "
    )
    ret = query_ucsc(command, database)

    # convert to dataframe
    df = pd.DataFrame.from_records(ret)
    df.columns = ["name", "accession_name2", "match"]
    df.set_index("name", inplace=True)

    # get best match
    df["match_max"] = df.groupby(df.index)["match"].max()
    df = df[df["match"] == df["match_max"]]

    # extract accessions
    # example: GCF_000090745.1_AnoCar2.0 -> GCF_000090745.1
    df = df["accession_name2"].str.extract(r"(GC[AF]_\d{9}\.\d+)")

    # GCA > GCF
    duplicates = df.index.duplicated(False)
    unique_df = df[~duplicates]
    dup_df = df[duplicates]
    filtered_dup_df = dup_df[dup_df[0].str[2] == "A"]
    df = pd.concat([unique_df, filtered_dup_df])

    # During testing, 93/217 genomes we assigned an accession ID
    accession_series = df[0]
    for name, acc in accession_series.items():
        if name not in genomes:
            continue
        genomes[name]["assembly_accession"] = acc

    return genomes


def add_accessions2(genomes: dict) -> dict:
    """
    Some genomes have their assembly accession in the 'sourceName' field.

    Updates the the genome dict "assembly_accession" field for each genome found.
    """
    re_acc = re.compile(r"GC[AF]_\d{9}\.\d+")
    for name in genomes:
        hit = re_acc.search(genomes[name]["sourceName"])
        if hit:
            genomes[name]["assembly_accession"] = hit.group()
    return genomes


def add_annotation_links(genomes):
    """
    identify the available annotation types for each genome.
    """
    # MySQL query
    command = (
        "SELECT TABLE_SCHEMA,TABLE_NAME "
        "FROM information_schema.tables "
        "WHERE table_name IN "
        "('ensGene', 'ncbiRefSeq', 'knownGene', 'refGene') "
    )
    ret = query_ucsc(command, database=None)

    for name, annot in ret:
        if name not in genomes:
            continue
        genomes[name]["annotations"].append(annot)

    return genomes


def query_ucsc(command: str, database: str = None) -> Generator:
    """
    Execute a single MySQL query on the UCSC database.
    Streams the output into a generator.
    """
    cnx = mysql.connector.connect(
        host="genome-mysql.soe.ucsc.edu",
        user="genome",
        port=3306,
        database=database,
    )
    try:
        cur = cnx.cursor(buffered=False, raw=False)
        cur.execute(command)
        while True:
            ret = cur.fetchone()
            if not ret:
                break
            yield ret
    finally:
        cnx.close()


def download_annotation(name, annot, genomes_dir, localname, n=None):
    """
    Download the extended genePred file from the UCSC MySQL database.
    Next convert this to a BED and GTF file.
    """
    check_ucsc_tools()

    out_dir = os.path.join(genomes_dir, localname)
    mkdir_p(out_dir)
    tmp_dir = mkdtemp(dir=out_dir)
    pred_file = f"{os.path.join(tmp_dir, localname)}.annotation.extended.gp"
    gtf_file = f"{os.path.join(out_dir, localname)}.annotation.gtf"
    bed_file = f"{os.path.join(out_dir, localname)}.annotation.bed"

    # MySQL query 1: get column names for this genePred
    command = f"SHOW COLUMNS FROM {annot};"
    cols = list(query_ucsc(command, database=name))

    # drop columns the UCSC tools cannot handle
    # see https://genome.ucsc.edu/FAQ/FAQformat.html#format9
    accepted_cols = [
        "geneName",
        "name",
        "chrom",
        "strand",
        "txStart",
        "txEnd",
        "cdsStart",
        "cdsEnd",
        "exonCount",
        "exonStarts",
        "exonEnds",
        "score",
        "name2",
        "cdsStartStat",
        "cdsEndStat",
        "exonFrames",
    ]
    cols = [c[0] for c in cols if c[0] in accepted_cols]
    cols = ",".join(cols)

    # MySQL query 2: download genePred
    command = f"SELECT {cols} FROM {annot};"
    if n:
        command = f"SELECT {cols} FROM {annot} LIMIT {n};"
    ret = query_ucsc(command, database=name)

    # clean up genePred
    df = pd.DataFrame.from_records(ret)
    for c in [8, 9, 14]:
        if c in df:
            df[c] = df[c].str.decode("utf-8")
    df.to_csv(pred_file, index=False, header=False, sep="\t")

    # convert genePred to GTF and BED
    cmd = "genePredToGtf -source=genomepy file {0} {1}"
    sp.check_call(cmd.format(pred_file, gtf_file), shell=True)
    cmd = "genePredToBed {0} {1}"
    sp.check_call(cmd.format(pred_file, bed_file), shell=True)
    rm_rf(tmp_dir)


@lock
@disk_cache.memoize(expire=cache_exp_genomes, tag="scrape_accession-ucsc")
def scrape_accession(htmlpath: str) -> str or None:
    """
    Attempt to scrape the assembly accession (`GCA_`/`GCF_`) from a genome's readme.html,
    or any linked NCBI assembly pages can also be scraped.

    Parameters
    ----------
    htmlpath: str
        path to the readme.tml on hgdownload.soe.ucsc.edu

    Returns
    ------
    str or None
        Assembly accession or 'na'
    """
    ucsc_url = f"https://hgdownload.soe.ucsc.edu/{htmlpath}"
    try:
        text = read_url(ucsc_url)
    except (UnicodeDecodeError, urllib.error.URLError):
        return None

    # example accessions: GCA_000004335.1 (ailMel1)
    # regex: GC[AF]_ = GCA_ or GCF_, \d = digit, \. = period
    accession_regex = re.compile(r"GC[AF]_\d{9}\.\d+")
    match = accession_regex.search(text)
    if match:
        return match.group(0)

    # Search for an assembly link at NCBI
    match = re.search(r"https?://www.ncbi.nlm.nih.gov/assembly/\d+", text)
    if match:
        ncbi_url = match.group(0)
        try:
            text = read_url(ncbi_url)
        except (UnicodeDecodeError, urllib.error.URLError):
            return None

        # retrieve valid assembly accessions.
        # contains additional info, such as '(latest)' or '(suppressed)'. Unused for now.
        valid_accessions = re.findall(r"assembly accession:.*?GC[AF]_.*?<", text)
        text = " ".join(valid_accessions)
        # this selects GCA > GCF if both are found
        match = accession_regex.search(text)
        if match:
            return match.group(0)



import pytest


def test_localprovider(local):
    assert local.name == "Local"
    assert local.genomes == {}


def test_genome_taxid(local):
    assert local.genome_taxid(None) is None


def test_assembly_accession(local):
    assert local.assembly_accession(None) is None


def test_search(local):
    result = local.search("???")
    with pytest.raises(StopIteration):
        assert next(result)


def test__genome_info_tuple(local):
    assert local._genome_info_tuple(None) is tuple()


def test__check_name(local):
    assert local._check_name(None) is None


def test_get_genome_download_link(local):
    target = "tests/data/sacCer3/sacCer3.fa"
    link = local.get_genome_download_link(target)
    assert link.endswith(target)

    with pytest.raises(FileNotFoundError):
        local.get_genome_download_link("this path does not exist")


def test_get_annotation_download_link(local):
    target = "tests/data/sacCer3/sacCer3.annotation.gtf"
    link = local.get_annotation_download_link(None, **{"path_to_annotation": target})
    assert link.endswith(target)

    with pytest.raises(FileNotFoundError):
        bad_path = "bad/path"
        local.get_annotation_download_link(None, **{"path_to_annotation": bad_path})

    with pytest.raises(TypeError):
        bad_ext = "tests/data/sacCer3/sacCer3.fa"
        local.get_annotation_download_link(None, **{"path_to_annotation": bad_ext})


def test_get_annotation_download_links(local):
    target = "tests/data/sacCer3/sacCer3.fa"
    expected = "tests/data/sacCer3/sacCer3.annotation.gtf"
    links = local.get_annotation_download_links(target)
    assert links[0].endswith(expected)



"""Annotation.sanitize method"""
from itertools import compress

from genomepy.annotation.utils import _check_property, write_annot
from genomepy.files import _open, update_readme


def _sanitize(self, match=True, filter=True, overwrite=False):  # noqa
    """
    Match the contigs names of the gene annotations to the genome's.

    First, match the contig names if possible.
    Second, remove contig names not found in the genome.
    Third, save the results and log this in the README.

    Parameters
    ----------
    match: bool, optional
        match annotation contig names to the genome contig names (default is True)
    filter: bool, optional
        remove annotation contig names not found in the genome contig names (default is True)
    overwrite: bool, optional
        update the annotation files on disk, and log this in the README (default is False).

    Returns
    -------
    Annotation class
        updated attributes
    """
    _check_property(self.genome_file, f"{self.name}.fa")

    cd = {}
    if match:
        cd = _match_contigs(self)

    mc = []
    if filter:
        mc = _filter_contigs(self)

    if overwrite:
        write_annot(self.gtf, self.annotation_gtf_file)
        write_annot(self.bed, self.annotation_bed_file)
        _document_sanitizing(self, cd, mc)


def _match_contigs(self):
    """
    Sometimes, the gene annotation contig names are different from the genome contig names.
    If the genome contains multiple headers, these may contain the annotation contig names.

    Match the gtf contig names to the genome contig names where possible.
    """
    # are there contigs in the annotation not found in the genome?
    missing_contigs = bool(set(self.annotation_contigs).difference(self.genome_contigs))

    if missing_contigs is False:
        return  # nothing to do

    # get the full headers from the fasta (e.g. '>chr1 chromosome1 note1 note2')
    headers = _full_genome_headers(self.genome_file)

    # filter for multi-head headers (whitespace delimited)
    multi_headed_contigs = [" " in h or "\t" in h for h in headers]

    if not any(multi_headed_contigs):
        return  # nothing to work with

    headers = [h.split() for h in compress(headers, multi_headed_contigs)]
    conversion_dict = _contig_conversion_dict(headers, self.annotation_contigs)

    self.gtf = self.gtf.replace({"seqname": conversion_dict})
    self.bed = self.bed.replace({"chrom": conversion_dict})
    return conversion_dict


def _filter_contigs(self):
    """
    Some tools throw a fit when the gene annotation contains
    contigs that are missing from the genome.

    Remove contigs not found in the genome.
    """
    # set of contigs in the genome that are not in the annotation
    missing_contigs = set(self.annotation_contigs).difference(self.genome_contigs)

    self.gtf = self.gtf[~self.gtf.seqname.isin(missing_contigs)]
    self.bed = self.bed[~self.bed.chrom.isin(missing_contigs)]
    return missing_contigs


def _document_sanitizing(self, cd, mc):
    _check_property(self.readme_file, "README.txt")
    status = ""
    extra_lines = []

    matched = bool(cd)
    if matched:
        status += f"{len(cd)} contigs were renamed. "

    filtered = bool(mc)
    if filtered:
        status += f"{len(mc)} contigs were removed (see below)."
        extra_lines = [
            "",
            "The following contigs were filtered out of the gene annotation:",
            f"{', '.join(mc)}",
        ]

    if not matched and not filtered:
        status = "no changes made."

    update_readme(self.readme_file, {"sanitized annotation": status}, extra_lines)


def _full_genome_headers(genome_file):
    headers = []
    with _open(genome_file) as fa:
        for line in fa:
            if line[0] == ">":
                headers.append(line.strip("> \n"))
    return headers


def _contig_conversion_dict(genome_headers, annotation_contigs):
    conversion_table = {}
    for header in genome_headers:
        for contig in annotation_contigs:
            if contig in header:
                conversion_table[contig] = header[0]
                break  # next header
    return conversion_table



import pytest

skip = False
if not skip:

    @pytest.fixture(scope="module", params=["primary_assembly", "toplevel"])
    def assembly(request):
        return request.param

    @pytest.fixture(scope="module", params=["98", None])
    def release_version(request):
        return request.param

    @pytest.fixture(scope="module", params=["hard", "soft", "unmasked"])
    def masking(request):
        return request.param

    def test_ensembl_genome_download_links(assembly, masking, release_version, ensembl):
        """Test Ensembl links with various options

        These genomes are hosted on ftp.ensembl.org

        Vertebrates are downloaded from HTTP.
        """
        mask = masking if masking != "unmasked" else "none"
        toplevel = False if assembly == "primary_assembly" else True
        version = release_version
        assert ensembl.get_genome_download_link(
            "GRCh38.p13", mask=mask, toplevel=toplevel, version=version
        )

    def test_ensemblgenomes_genome_download_links(masking, ensembl):
        """Test Ensembl FTP links for various genomes

        These genomes are hosted on ftp.ensemblgenomes.org.
        """
        mask = masking if masking != "unmasked" else "none"
        for genome in ["Amel_HAv3.1", "ASM23943v1"]:
            assert ensembl.get_genome_download_link(
                genome, mask=mask, **{"version": 55}
            )

    def test_ucsc_genome_download_links(masking, ucsc):
        """Test UCSC HTTP links for various genomes

        Also test masking (unmasked should be ignored)."""
        for genome in ["sacCer3", "hg38"]:
            assert ucsc.get_genome_download_link(genome, mask=masking)

    def test_ncbi_genome_download_links(masking, ncbi):
        """Test NCBI HTTPS links for various genomes

        Also test masking (should be ignored).

        These genomes are hosted on ftp://ftp.ncbi.nlm.nih.gov."""
        for genome in ["Charlie1.0", "GRCh38.p13"]:
            assert ncbi.get_genome_download_link(genome, mask=masking)



import re
from typing import List

from loguru import logger

from genomepy.exceptions import GenomeDownloadError
from genomepy.files import get_file_info
from genomepy.online import read_url
from genomepy.providers.base import BaseProvider
from genomepy.utils import get_localname


class UrlProvider(BaseProvider):
    """
    URL genome provider.

    Simply download a genome directly through a url.
    """

    name = "URL"
    _cli_install_options = {
        "to_annotation": {
            "long": "to-annotation",
            "help": "link to the annotation file, required if this is not in the same directory as the fasta file",
            "default": None,
        },
    }

    def __init__(self):
        self.genomes = {}

    @staticmethod
    def ping():
        return True

    def genome_taxid(self, name):
        return

    def assembly_accession(self, name):
        return

    def search(self, term, size=False):
        """return an empty generator,
        same as if no genomes were found at the other providers"""
        yield from ()

    def _genome_info_tuple(self, name, size=False):
        return tuple()

    def _check_name(self, name):
        """check if genome name can be found for provider"""
        return name

    def get_genome_download_link(self, url, mask=None, **kwargs):
        return url

    def get_annotation_download_link(self, name: str, **kwargs) -> str:
        """
        Return a functional annotation download link.

        Parameters
        ----------
        name : str
            genome name
        **kwargs: dict, optional:
            to_annotation : direct URL to the gene annotation

        Returns
        -------
        str
            http/ftp link

        Raises
        ------
        GenomeDownloadError
            if no functional link was found
        """
        link = kwargs.get("to_annotation")
        if link:
            ext = get_file_info(link)[0]
            if ext not in [".gtf", ".gff", ".gff3", ".bed"]:
                raise TypeError(
                    "Only (gzipped) gtf, gff and bed files are supported.\n"
                )
            return link

        links = self.get_annotation_download_links(name)
        if links:
            return links[0]

        raise GenomeDownloadError(
            f"No gene annotations found for {get_localname(name)}.\n"
        )

    def get_annotation_download_links(self, name: str, **kwargs) -> List[str]:
        """
        Retrieve functioning gene annotation download link(s).

        If provided, check if the annotation url links to a supported file type (gtf/gff3/bed).
        Else try to find an annotation in the same location as the genome url.

        Parameters
        ----------
        name : str
            genome name

        Returns
        -------
        list
            http/ftp link(s)
        """
        # name = url to genome
        return search_url_for_annotations(name)


def search_url_for_annotations(url: str) -> list:
    """Attempts to find gtf or gff3 files in the same location as the genome url"""
    name = get_localname(url)

    urldir = url[: url.rfind("/")]
    logger.info(
        "You have requested the gene annotation to be downloaded. "
        "Genomepy will check the remote directory: "
        f"{urldir} for annotation files..."
    )

    # try to find a GTF or GFF3 file
    dirty_list = read_url(urldir).split("\n")
    fnames = fuzzy_annotation_search(name, dirty_list)
    links = [urldir + "/" + fname for fname in fnames]
    if not links:
        logger.warning(
            "Could not parse the remote directory. "
            "Please supply a URL using --URL-to-annotation.\n"
        )
    return links


def fuzzy_annotation_search(search_name, search_list):
    """Returns all files containing both name and an annotation extension"""
    hits = []
    for ext in ["gtf", "gff"]:
        # .*? = non greedy filler. 3? = optional 3 (for gff3). (\.gz)? = optional .gz
        expr = rf"{search_name}.*?\.{ext}3?(\.gz)?"  # noqa: W605
        for line in search_list:
            hit = re.search(expr, line, flags=re.IGNORECASE)
            if hit:
                hits.append(hit[0])
    return hits



"""Utility functions with internet connections"""
import socket
import urllib.error
from ftplib import FTP, all_errors, error_temp
from time import sleep
from typing import Tuple
from urllib.request import urlopen

import requests
from tqdm.auto import tqdm

from genomepy.exceptions import GenomeDownloadError


def download_file(url, filename) -> str:
    """
    Helper method handling downloading large files from `url` to `filename`.

    Parameters
    ----------
    url : str
        download target url
    filename : str
        file to download to

    Returns
    -------
    str
        filename
    """

    def decorated_pbar(total):
        """Displays a progress bar with download speeds in MB/s."""
        return tqdm(
            desc="Download",
            unit_scale=True,
            unit_divisor=1024,
            total=total,
            unit="B",
        )

    def write_n_update_pbar(data):
        pbar.update(len(data))
        f.write(data)

    if url.startswith("ftp"):
        ftp, target = connect_ftp_link(url)
        file_size = ftp.size(target)
        with open(filename, "wb") as f:
            pbar = decorated_pbar(file_size)

            ftp.retrbinary(f"RETR {target}", write_n_update_pbar)
            ftp.quit()  # logout

    else:
        r = requests.get(url, stream=True)
        file_size = int(r.headers.get("Content-Length", 0))
        with open(filename, "wb") as f:
            pbar = decorated_pbar(file_size)

            for chunk in r.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    write_n_update_pbar(chunk)

    pbar.close()  # stop updating pbar
    return filename


def connect_ftp_link(link, timeout=None) -> Tuple[FTP, str]:
    """
    Anonymous login to ftp.

    Accepts link in the form of ftp://ftp.name.domain/...
    and ftp.name.domain/...

    Parameters
    ----------
    link : str
        FTP link
    timeout : int, optional
        number of idle seconds before the connection closes

    Returns
    -------
    tuple
        ftp: FTP
            object with connection established
        target: str
            target file
    """
    link = link.replace("ftp://", "")
    host, target = link.split("/", 1)
    try:
        ftp = FTP(host, timeout=timeout)
    except socket.gaierror:
        raise GenomeDownloadError(f"FTP host not found: {host}")
    ftp.login()
    return ftp, target


def read_url(url):
    """Read a text-based URL."""
    response = urlopen(url)
    data = response.read()
    text = data.decode("utf-8")
    return text


def retry(func, tries, *args, **kwargs):
    """
    Retry functions with potential connection errors.

    args and kwargs are passed to func.
    """
    for n in range(tries):
        try:
            answer = func(*args, **kwargs)
            return answer
        except (urllib.error.HTTPError, error_temp):
            # HTTPError 404: URL not found
            # FTP error_temp 450: file not found
            return
        except all_errors + (urllib.error.URLError, socket.timeout):
            # connection errors: wait and try again
            if n < tries - 1:
                sleep(1)


def check_url(url, max_tries=1, timeout=15) -> bool:
    """Check if URL works. Returns bool"""

    def _check_url(_url, _timeout):
        if _url.startswith("ftp"):
            ftp, target = connect_ftp_link(_url, timeout=_timeout)
            listing = retry(ftp.nlst, 1, target)
            ftp.quit()  # logout
            if listing:
                return True
        else:
            ret = urlopen(_url, timeout=_timeout)
            if ret.getcode() == 200:
                return True

        return False

    return retry(_check_url, max_tries, url, timeout)



import os
import re
import urllib.error
from typing import List
from urllib.request import urlopen

import pandas as pd
from loguru import logger
from tqdm.auto import tqdm

from genomepy.caching import cache_exp_genomes, disk_cache, lock
from genomepy.exceptions import GenomeDownloadError
from genomepy.online import check_url, read_url
from genomepy.providers.base import BaseProvider
from genomepy.utils import mkdir_p, safe

ASM_FORMAT = [
    "Sequence-Name",
    "Sequence-Role",
    "Assigned-Molecule",
    "Assigned-Molecule-Location/Type",
    "GenBank-Accn",
    "Relationship",
    "RefSeq-Accn",
    "Assembly-Unit",
    "Sequence-Length",
    "UCSC-style-name",
]


class NcbiProvider(BaseProvider):
    """
    NCBI genome provider.

    Uses the assembly reports page to search and list genomes.
    """

    name = "NCBI"
    accession_fields = ["assembly_accession", "gbrs_paired_asm"]
    taxid_fields = ["species_taxid", "taxid"]
    description_fields = [
        "submitter",
        "organism_name",
        "assembly_accession",
        "gbrs_paired_asm",
        "paired_asm_comp",
    ]
    _cli_install_options = {}
    _url = "https://ftp.ncbi.nlm.nih.gov/genomes/ASSEMBLY_REPORTS/"

    def __init__(self):
        self._provider_status()
        # Populate on init, so that methods can be cached
        self.genomes = get_genomes(self._url)

    @staticmethod
    def ping():
        """Can the provider be reached?"""
        url_online = bool(check_url("https://ftp.ncbi.nlm.nih.gov/genomes/"))
        return url_online

    def _genome_info_tuple(self, name, size=False):
        """tuple with assembly metadata"""
        accession = self.assembly_accession(name)
        taxid = self.genome_taxid(name)
        annotations = bool(self.annotation_links(name))
        species = self.genomes[name].get("organism_name")
        other = self.genomes[name].get("submitter")
        if size:
            length = get_genome_size(accession)
            return name, accession, taxid, annotations, species, length, other
        return name, accession, taxid, annotations, species, other

    def get_genome_download_link(self, name, mask="soft", **kwargs):
        """
        Return NCBI ftp link to top-level genome sequence

        Parameters
        ----------
        name : str
            Genome name. Current implementation will fail if exact
            name is not found.

        mask : str , optional
            Masking level. Options: soft, hard or none. Default is soft.

        Returns
        ------
        str with the http/ftp download link.
        """
        # only soft masked genomes available. can be (un)masked in _post_process_download
        link = self._ftp_or_html_link(name, file_suffix="_genomic.fna.gz")

        if link:
            return link

        raise GenomeDownloadError(
            f"Could not download genome {name} from {self.name}.\n"
            "URL is broken. Select another genome or provider.\n"
            f"Broken URL: {link}"
        )

    def _post_process_download(self, name, fname, out_dir, mask="soft"):
        """
        Replace accessions with sequence names in fasta file.

        Applies masking.

        Parameters
        ----------
        name : str
            NCBI genome name

        fname : str
            file path to the genome fasta

        out_dir : str
            output directory

        mask : str , optional
            masking level: soft/hard/none, default=soft
        """
        # (down)load the assembly report
        asm_fname = os.path.join(out_dir, "assembly_report.txt")
        if not os.path.exists(asm_fname):
            # save assembly report
            download_assembly_report(self.assembly_accession(name), asm_fname)
        asm_report = pd.read_csv(
            asm_fname, sep="\t", comment="#", header=None, dtype=str
        )

        # create mapping of chromosome accessions to chromosome names
        tr = asm_report.set_index(6)[0].to_dict()

        # mask sequence if required
        if mask == "soft":

            def mask_cmd(txt):
                return txt

        elif mask == "hard":
            logger.info("NCBI genomes are softmasked by default. Hard masking...")
            lower_case_nucleotides = re.compile("[actg]")

            def mask_cmd(txt):
                return lower_case_nucleotides.sub("N", txt)

        else:
            logger.info("NCBI genomes are softmasked by default. Unmasking...")

            def mask_cmd(txt):
                return txt.upper()

        # apply mapping and masking
        old_fname = os.path.join(
            os.path.dirname(fname), f"original_{os.path.basename(fname)}"
        )
        os.rename(fname, old_fname)
        with open(old_fname) as old, open(fname, "w") as new:
            for line in tqdm(
                old, desc="Processing NCBI Fasta", unit_scale=1, unit=" lines"
            ):
                if line[0] == ">":
                    desc = line[1:].strip()
                    name = desc.split(" ")[0]
                    new.write(f">{tr.get(name, name)} {desc}\n")
                else:
                    new.write(mask_cmd(line))

    def get_annotation_download_links(self, name: str, **kwargs) -> List[str]:
        """
        Retrieve functioning gene annotation download link(s).

        Parse and test the link to the NCBI annotation file.

        Parameters
        ----------
        name : str
            genome name

        Returns
        -------
        list
            http/ftp link(s)
        """
        link = self._ftp_or_html_link(name, file_suffix="_genomic.gff.gz")
        return [link] if link else []

    def _ftp_or_html_link(self, name, file_suffix, skip_check=False):
        """
        NCBI's files are accessible over FTP and HTTPS
        Try HTTPS first and return the first functioning link
        """
        genome = self.genomes[safe(name)]
        ftp_link = genome["ftp_path"]
        html_link = ftp_link.replace("ftp://", "https://")
        for link in [html_link, ftp_link]:
            link += "/" + link.split("/")[-1] + file_suffix

            if skip_check or check_url(link, max_tries=2, timeout=10):
                return link


@lock
@disk_cache.memoize(expire=cache_exp_genomes, tag="get_genomes-ncbi")
def get_genomes(assembly_url):
    """Parse genomes from assembly summary txt files."""
    logger.info("Downloading assembly summaries from NCBI, this will take a while...")

    def load_summary(url):
        """
        lazy loading of the url so we can parse while downloading
        """
        for row in urlopen(url):
            yield row

    genomes = {}
    # order is important as asm_name can repeat (overwriting the older name)
    names = [
        "assembly_summary_genbank_historical.txt",
        "assembly_summary_refseq_historical.txt",
        "assembly_summary_genbank.txt",
        "assembly_summary_refseq.txt",
    ]
    # filter summaries to these keys (to reduce the size of the cached data)
    summary_keys_to_keep = [
        0,  # 'assembly_accession',
        5,  # 'taxid',
        6,  # 'species_taxid',
        7,  # 'organism_name',
        16,  # 'submitter',
        17,  # 'gbrs_paired_asm',
        18,  # 'paired_asm_comp',
        19,  # 'ftp_path',
    ]
    for fname in names:
        lines = load_summary(f"{assembly_url}/{fname}")
        # line 0 = comment
        _ = next(lines)
        # line 1 = header
        header = next(lines).decode("utf-8").strip("# ").strip("\n").split("\t")
        header = [header[n] for n in summary_keys_to_keep]
        for line in tqdm(lines, desc=fname[17:-4], unit_scale=1, unit=" genomes"):
            line = line.decode("utf-8").strip("\n").split("\t")
            name = safe(line[15])  # overwrites older asm_names
            if line[19] != "na":  # ftp_path must exist
                line = [line[n] for n in summary_keys_to_keep]
                genomes[name] = dict(zip(header, line))
    return genomes


def download_assembly_report(acc: str, fname: str = None, quiet=False):
    """
    Retrieve the NCBI assembly report.

    Returns the assembly_report as a pandas DataFrame if fname is not specified.

    Parameters
    ----------
    acc : str
        Assembly accession (GCA or GCF)
    fname : str, optional
        Save assembly_report to this filename.
    quiet : bool, optional
        Silence warnings.

    Returns
    -------
    pandas.DataFrame
        NCBI assembly report.
    """
    msg = "Could not download the assembly report from NCBI. "
    if not isinstance(acc, str) or not acc.startswith(("GCA", "GCF")):
        if not quiet:
            logger.warning(msg)
        return None
    assembly_report = _assembly_report_url(acc)
    if assembly_report is None:
        if not quiet:
            logger.warning(msg + f"Assembly accession '{acc}' not found.")
        return None
    asm_report = pd.read_csv(
        assembly_report, sep="\t", comment="#", names=ASM_FORMAT, dtype=str
    )

    if fname:
        mkdir_p(os.path.dirname(fname))
        asm_report.to_csv(fname, sep="\t", index=False)
    else:
        return asm_report


def _assembly_report_url(acc: str) -> str or None:
    ftp_basedir = "https://ftp.ncbi.nlm.nih.gov/genomes/all"
    asm_dir = f"{ftp_basedir}/{acc[0:3]}/{acc[4:7]}/{acc[7:10]}/{acc[10:13]}/"
    try:
        text = read_url(asm_dir)
    except urllib.error.HTTPError:
        return None

    # try to find exact accession patch level
    hits = re.findall(acc + '_.*/"', text)  # len 0-1
    if len(hits) == 0:
        # patch level not found, pick the closest instead
        hits = re.findall(acc + '.*/"', text)
        available_accessions = ["_".join(h.split("_")[0:2]) for h in hits]
        closest = _closest_patch_lvl(acc, available_accessions)
        hits = [h for h in hits if closest + "_" in h]  # len 0-1
    if len(hits) == 0:
        return None

    name = hits[0][:-2]
    assembly_report = asm_dir + name + "/" + name + "_assembly_report.txt"
    return assembly_report


def _closest_patch_lvl(reference: str, targets: list) -> str:
    ref_patch = _patch_lvl(reference)
    closest = ""
    closest_dist = 999
    for target in targets:
        tgt_patch = _patch_lvl(target)
        dist = abs(tgt_patch - 0.1 - ref_patch)  # tiebreaker: newer > older patches
        if dist < closest_dist:
            closest = target
            closest_dist = dist
    return closest


def _patch_lvl(acc: str) -> int:
    """str(GCA_000000000.6) -> int(6)"""
    return int(acc.split(".")[1]) if "." in acc else 0


def get_genome_size(acc):
    length = -1
    df = download_assembly_report(acc, quiet=True)
    if df is not None:
        length = df["Sequence-Length"].astype(int).sum()
    return length



"""BaseProvider class, the parent of the provider classes"""
import gzip
import os
import re
import shutil
import subprocess as sp
import time
from tempfile import TemporaryDirectory, mkdtemp
from typing import Iterator, List
from urllib.request import urlopen

import pandas as pd
from loguru import logger

from genomepy.__about__ import __version__
from genomepy.annotation.utils import read_annot, write_annot
from genomepy.exceptions import GenomeDownloadError
from genomepy.files import extract_archive, get_file_info, update_readme
from genomepy.online import download_file
from genomepy.utils import (
    check_ucsc_tools,
    get_genomes_dir,
    get_localname,
    mkdir_p,
    rm_rf,
    safe,
)


class BaseProvider:
    """
    Provider base class.
    """

    # class variables set by child classes:
    name = None
    "Name of this provider."
    genomes = {}
    "Dictionary with assembly names as key and assembly metadata dictionary as value."
    accession_fields = []
    "Metadata fields that (can) contain the assembly's accession ID."
    taxid_fields = []
    "Metadata fields that (can) contain the assembly's taxonomy ID."
    description_fields = []
    "Metadata fields with assembly related info."
    _cli_install_options = {}
    _url = None

    def __hash__(self):
        return hash(str(self.__class__))

    @staticmethod
    def ping() -> bool:
        """Can the provider be reached?"""
        raise NotImplementedError()

    def _provider_status(self):
        """check if provider is online"""
        if not self.ping():
            raise ConnectionError(f"{self.name} appears to be offline.")

    def _check_name(self, name):
        """check if genome name can be found for provider"""
        name = safe(name)
        if name in self.genomes:
            return name

        raise GenomeDownloadError(
            f"Could not download genome {name} from {self.name}.\n\n"
            "Check for typos or try\n"
            f"  genomepy search {name} -p {self.name}"
        )

    def _genome_info_tuple(self, name, size=False):
        """tuple with assembly metadata"""
        raise NotImplementedError()

    def list_available_genomes(self, size=False):
        """
        List all available genomes.

        Yields
        ------
        genomes : list of tuples
            tuples with assembly name, accession, scientific_name, taxonomy id and description
        size : bool, optional
            Show absolute genome size.
        """
        for name in self.genomes.keys():
            yield self._genome_info_tuple(name, size)

    def genome_taxid(self, name: str) -> int:
        """
        Return the genome taxonomy ID for a genome.

        Parameters
        ----------
        name: str
            genome name

        Returns
        ------
        int
            Genome Taxonomy identifier
        """
        for field in self.taxid_fields:
            tid = str(self.genomes[name].get(field))
            if tid.isdigit():
                return int(tid)

    def assembly_accession(self, name: str) -> str or None:
        """
        Return the assembly accession number (GCA* or GCF*) for a genome.

        Parameters
        ----------
        name: str
            genome name

        Returns
        ------
        str
            Assembly accession number
        """
        for field in self.accession_fields:
            accession = str(self.genomes[name].get(field))
            if accession.startswith(("GCA", "GCF")):
                return accession

    def annotation_links(self, name: str, **kwargs) -> List[str]:
        """
        Return available gene annotation links (http/ftp) for a genome

        Parameters
        ----------
        name: str
            genome name

        Returns
        ------
        list
            Gene annotation links
        """
        if "annotations" not in self.genomes[safe(name)]:
            links = self.get_annotation_download_links(name, **kwargs)
            self.genomes[safe(name)]["annotations"] = links
        return self.genomes[safe(name)]["annotations"]

    def get_genome_download_link(self, name, mask="soft", **kwargs):
        raise NotImplementedError()

    def download_genome(
        self,
        name: str,
        genomes_dir: str = None,
        localname: str = None,
        mask: str = "soft",
        **kwargs,
    ):
        """
        Download a (gzipped) genome file to a specific directory

        Parameters
        ----------
        name : str
            Genome / species name

        genomes_dir : str , optional
            Directory to install genome

        localname : str , optional
            Custom name for your genome

        mask: str , optional
            Masking, soft, hard or none (all other strings)
        """
        name = self._check_name(name)
        link = self.get_genome_download_link(name, mask=mask, **kwargs)

        localname = get_localname(name, localname)
        genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)
        out_dir = os.path.join(genomes_dir, localname)
        mkdir_p(out_dir)

        logger.info(f"Downloading genome from {self.name}. Target URL: {link}...")

        # download to tmp dir. Move genome on completion.
        # tmp dir is in genome_dir to prevent moving the genome between disks
        get_file = shutil.copyfile if os.path.exists(link) else download_file
        with TemporaryDirectory(dir=out_dir) as tmp_dir:
            tmp_fname = os.path.join(tmp_dir, link.split("/")[-1])
            fname = os.path.join(tmp_dir, f"{localname}.fa")

            get_file(link, tmp_fname)
            logger.info("Genome download successful, starting post processing...")

            # unzip genome
            _, is_compressed = get_file_info(link)
            if is_compressed:
                extract_archive(tmp_fname, outfile=fname, concat=True)
            else:
                shutil.move(tmp_fname, fname)

            # process genome (e.g. masking)
            if hasattr(self, "_post_process_download"):
                self._post_process_download(
                    name=name, fname=fname, out_dir=out_dir, mask=mask
                )

            # transfer the genome from the tmpdir to the genome_dir
            src = fname
            dst = os.path.join(out_dir, f"{localname}.fa")
            shutil.move(src, dst)

        logger.info("name: {}".format(name))
        logger.info("local name: {}".format(localname))
        logger.info("fasta: {}".format(dst))

        # Create readme with information
        readme = os.path.join(genomes_dir, localname, "README.txt")
        asm_acc = self.assembly_accession(name)
        tax_id = self.genome_taxid(name)
        metadata = {
            "name": localname,
            "provider": self.name,
            "original name": name,
            "original filename": os.path.split(link)[-1],
            "assembly_accession": asm_acc if asm_acc else "na",
            "tax_id": tax_id if tax_id else "na",
            "mask": mask,
            "genome url": link,
            "genomepy version": __version__,
            "date": time.strftime("%Y-%m-%d %H:%M:%S"),
        }
        update_readme(readme, metadata)

    def get_annotation_download_links(self, name, **kwargs):
        """
        Retrieve functioning gene annotation download link(s).

        Parameters
        ----------
        name : str
            genome name
        **kwargs: dict, optional:
            provider specific options.

        Returns
        -------
        list
            http/ftp link(s)
        """
        raise NotImplementedError()

    def get_annotation_download_link(self, name: str, **kwargs) -> str:
        """
        Return a functional annotation download link.

        Parameters
        ----------
        name : str
            genome name

        Returns
        -------
        str
            http/ftp link

        Raises
        ------
        GenomeDownloadError
            if no functional link was found
        """
        links = self.annotation_links(name, **kwargs)
        if links:
            return links[0]
        raise GenomeDownloadError(
            f"No gene annotations found for {name} on {self.name}.\n"
            "Check for typos or try\n"
            f"  genomepy search {name} -p {self.name}"
        )

    def download_annotation(self, name, genomes_dir=None, localname=None, **kwargs):
        """
        Download annotation file to to a specific directory

        Parameters
        ----------
        name : str
            Genome / species name

        genomes_dir : str , optional
            Directory to install annotation

        localname : str , optional
            Custom name for your genome
        """
        name = self._check_name(name)
        link = self.get_annotation_download_link(name, **kwargs)

        localname = get_localname(name, localname)
        genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)

        logger.info(f"Downloading annotation from {self.name}. Target URL: {link}...")
        try:
            download_annotation(genomes_dir, link, localname)
            logger.info("Annotation download successful")
        except Exception as e:
            raise GenomeDownloadError(
                f"An error occurred while installing the gene annotation for {name} from {self.name}.\n"
                "If you think the annotation should be there, please file a bug report at: "
                "https://github.com/vanheeringen-lab/genomepy/issues\n\n"
                f"Error: {e.args[0]}"
            )

        # Add annotation URL to readme
        readme = os.path.join(genomes_dir, localname, "README.txt")
        update_readme(readme, updated_metadata={"annotation url": link})

    def _search_text(self, term: str) -> Iterator[str]:
        """check if search term is found in the provider's genome name or description field(s)"""
        # multiple search terms: order doesn't matter
        if " " in term:
            term = "".join([f"(?=.*{t})" for t in term.split()])

        pattern = re.compile(term, re.I)  # case insensitive
        for name, metadata in self.genomes.items():
            texts = [name] + [str(metadata[f]) for f in self.description_fields]
            if any(pattern.search(t) for t in texts):
                yield name

    def _search_accession(self, term: str) -> Iterator[str]:
        """check if search term is found in the provider's accession field(s)"""
        # cut off prefix (GCA_/GCF_) and suffix (version numbers, e.g. '.3')
        term = term[4:].split(".")[0]
        for name, metadata in self.genomes.items():
            if any(term in str(metadata[f]) for f in self.accession_fields):
                yield name

    def _search_taxonomy(self, term: str) -> Iterator[str]:
        """check if search term is the start of the provider's taxonomy field(s)"""
        for name, metadata in self.genomes.items():
            if any(
                str(metadata[f]).strip().startswith(term) for f in self.taxid_fields
            ):
                yield name

    def search(self, term: str or int, size=False):
        """
        Search for term in genome names, descriptions and taxonomy ID.

        The search is case-insensitive.

        Parameters
        ----------
        term : str, int
            Search term, case-insensitive.
            Can be (part of) an assembly name (e.g. hg38),
            scientific name (Danio rerio) or assembly
            accession (`GCA_000146045`/`GCF_`),
            or a taxonomy id (7227).
        size : bool, optional
            Show absolute genome size.

        Yields
        ------
        tuples with name and metadata
        """
        term = str(term).strip()

        if term.lower().startswith(("gca_", "gcf_")):
            search_function = self._search_accession
        elif term.isdigit():
            search_function = self._search_taxonomy
        else:
            search_function = self._search_text

        for name in search_function(term):
            yield self._genome_info_tuple(name, size)

    def head_annotation(self, name: str, genomes_dir=None, n: int = 5, **kwargs):
        """
        Download the first n lines of the annotation.

        The first line of the GTF is printed for review
        (of the gene_name field, for instance).

        Parameters
        ----------
        name : str
            genome name
        genomes_dir : str, optional
            genomes directory to install the annotation in.
        n : int, optional
            download the annotation for n genes.
        """
        name = self._check_name(name)
        link = self.get_annotation_download_link(name, **kwargs)

        localname = f"{name}_head"
        genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)

        fpath = os.path.join(genomes_dir, localname, f"{localname}.annotation.gtf")
        download_annotation(genomes_dir, link, localname, n=n)

        logger.info(self.name)
        m = 0
        with open(fpath) as f:
            for line in f:
                line = line.strip()
                if line and line[0] != "#":
                    print(line)
                    m += 1
                if m == n:
                    break


def download_annotation(genomes_dir, annot_url, localname, n=None):
    """download annotation file, convert to intermediate file and generate output files"""

    check_ucsc_tools()

    # create output directory if missing
    out_dir = os.path.join(genomes_dir, localname)
    mkdir_p(out_dir)

    # download to tmp dir. Move genome on completion.
    # tmp dir is in genome_dir to prevent moving the genome between disks
    tmp_dir = mkdtemp(dir=out_dir)
    ext, is_compressed = get_file_info(annot_url)

    annot_file = os.path.join(tmp_dir, localname + ".annotation" + ext)
    tmp_annot_file = os.path.join(tmp_dir, annot_url.split("/")[-1])
    get_file = shutil.copyfile if os.path.exists(annot_url) else download_file
    if n is None:
        get_file(annot_url, tmp_annot_file)
    else:
        download_head(annot_url, tmp_annot_file, n)
        is_compressed = False

    # unzip input file (if needed)
    if is_compressed:
        annot_file = extract_archive(tmp_annot_file, outfile=annot_file)
    else:
        shutil.move(tmp_annot_file, annot_file)

    # generate intermediate file (GenePred)
    pred_file = annot_file.replace(ext, ".gp")
    if "bed" in ext:
        cmd = "bedToGenePred {0} {1}"
    elif "gff" in ext:
        # example annotation: GRCh38.p12 from NCBI
        cmd = "gff3ToGenePred -useName -warnAndContinue {0} {1}"
    elif "gtf" in ext:
        cmd = "gtfToGenePred -genePredExt -allErrors -ignoreGroupsWithoutExons {0} {1}"
    elif "txt" in ext:
        # UCSC annotations only
        with open(annot_file) as f:
            cols = f.readline().split("\t")

        # extract the genePred format columns
        start_col = 1
        for i, col in enumerate(cols):
            if col in ["+", "-"]:
                start_col = i - 1
                break
        end_col = start_col + 10
        cmd = (
            f"""cat {{0}} | cut -f {start_col}-{end_col} | """
            # knownGene.txt.gz has spotty fields, this replaces non-integer fields with zeroes
            + """awk 'BEGIN {{FS=OFS="\t"}} !($11 ~ /^[0-9]+$/) {{$11="0"}}1' > {1}"""
        )
    else:
        raise TypeError(f"file type extension {ext} not recognized!")

    if n is None and "gencode" in annot_url:
        rename_contigs(annot_file)

    sp.check_call(cmd.format(annot_file, pred_file), shell=True)

    # generate gzipped gtf file (if required)
    gtf_file = annot_file.replace(ext, ".gtf")
    if "gtf" not in ext:
        cmd = "genePredToGtf -source=genomepy file {0} {1}"
        sp.check_call(cmd.format(pred_file, gtf_file), shell=True)

    # generate gzipped bed file (if required)
    bed_file = annot_file.replace(ext, ".bed")
    if "bed" not in ext:
        cmd = "genePredToBed {0} {1}"
        sp.check_call(cmd.format(pred_file, bed_file), shell=True)

    # transfer the files from the tmpdir to the genome_dir
    for f in [gtf_file, bed_file]:
        src = f
        dst = os.path.join(out_dir, os.path.basename(f))
        shutil.move(src, dst)
    rm_rf(tmp_dir)


def download_head(annot_url, annot_file, n: int = 5):
    """
    Download the first n lines of a (gzipped) file.
    Comment lines are downloaded but do not count towards the line limit.

    Parameters
    ----------
    annot_url : str
        url to the file
    annot_file : str
        output filename
    n : int, optional
        number of lines to download
    """
    res = urlopen(annot_url)
    if annot_url.endswith(".gz"):
        res = gzip.GzipFile(fileobj=res)
    m = 0
    with open(annot_file, "w") as f:
        for line in res:
            line = line.decode("utf-8")
            f.write(line)
            if line[0] != "#":
                m += 1
            # add a few extra lines to the intermediate file
            if m == n + 2:
                break


def rename_contigs(annot_file):
    genome_dir = os.path.dirname(os.path.dirname(annot_file))
    asm_report = os.path.join(genome_dir, "assembly_report.txt")
    gencode2ucsc = pd.read_csv(
        asm_report,
        sep="\t",
        comment="#",
        usecols=["GenBank-Accn", "UCSC-style-name"],
        dtype=str,
    )
    gtf = read_annot(annot_file)

    # use the UCSC names for the scaffolds
    newgtf = gtf.merge(
        gencode2ucsc, left_on="seqname", right_on="GenBank-Accn", how="left"
    )
    newgtf["seqname"] = newgtf["UCSC-style-name"].mask(pd.isnull, newgtf["seqname"])
    newgtf.drop(columns=["GenBank-Accn", "UCSC-style-name"], inplace=True)

    # overwrite the raw GTF
    write_annot(newgtf, annot_file)



import os
from tempfile import TemporaryDirectory

import pytest

import genomepy.files
import genomepy.utils
from tests import travis
from tests.conftest import validate_annot


def test_baseprovider(base):
    assert base.name is None


def test__check_name(ucsc):
    ucsc._check_name("ailMel1")

    with pytest.raises(genomepy.exceptions.GenomeDownloadError):
        ucsc._check_name("not_a_real_genome")


def test__genome_info_tuple(base):
    with pytest.raises(NotImplementedError):
        base._genome_info_tuple(None)


def test_list_available_genomes(base):
    assert list(base.list_available_genomes()) == []


def test_genome_taxid(ucsc):
    taxid = ucsc.genome_taxid("ailMel1")
    assert taxid == 9646


def test_assembly_accession(ncbi):
    accession = ncbi.assembly_accession("AilMel_1.0")
    assert "000004335" in accession


def test_annotation_links(ncbi):
    # most genomes: 1 annotation
    links = ncbi.annotation_links("AilMel_1.0")
    expected = (
        "ftp.ncbi.nlm.nih.gov/genomes/all/GCA/000/004/335/"
        "GCA_000004335.1_AilMel_1.0/GCA_000004335.1_AilMel_1.0_genomic.gff.gz"
    )
    assert len(links) == 1
    assert links[0].endswith(expected)

    # some genomes: no annotation
    links = ncbi.annotation_links("Oryza_glaberrima_V1")
    assert links == []


def get_genome_download_link(base):
    with pytest.raises(NotImplementedError):
        base.get_genome_download_link()


@pytest.mark.skipif(not travis, reason="slow")
def test_download_genome(
    ucsc,
    name="sacCer3",
    localname="my_genome",
    mask="soft",
):
    out_dir = os.getcwd()

    with TemporaryDirectory(dir=out_dir) as tmpdir:
        ucsc.download_genome(
            name,
            genomes_dir=tmpdir,
            localname=localname,
            mask=mask,
        )

        genome = os.path.join(tmpdir, localname, localname + ".fa")
        assert os.path.exists(genome)

        readme = os.path.join(os.path.dirname(genome), "README.txt")
        with open(readme) as f:
            metadata = {}
            for line in f.readlines():
                vals = line.strip().split(":")
                metadata[vals[0].strip()] = (":".join(vals[1:])).strip()

        assert metadata["name"] == localname
        assert metadata["mask"] == mask


def test_get_annotation_download_links(base):
    with pytest.raises(NotImplementedError):
        base.get_annotation_download_links(None)


def test_get_annotation_download_link(ncbi):
    link = ncbi.get_annotation_download_link("AilMel_1.0")
    expected = (
        "ftp.ncbi.nlm.nih.gov/genomes/all/GCA/000/004/335/"
        "GCA_000004335.1_AilMel_1.0/GCA_000004335.1_AilMel_1.0_genomic.gff.gz"
    )
    assert link.endswith(expected)


# @pytest.mark.skipif(not travis, reason="slow")
# def test__download_annotation(base):
#     out_dir = os.getcwd()
#     localname = "my_annot"
#
#     annot_url = "https://www.google.com"
#     with pytest.raises(TypeError), TemporaryDirectory(dir=out_dir) as tmpdir:
#         base._download_annotation(
#             genomes_dir=tmpdir, annot_url=annot_url, localname=localname
#         )
#
#     annot_url = "http://hgdownload.soe.ucsc.edu/goldenPath/ailMel1/bigZips/genes/ailMel1.ensGene.gtf.gz"
#     with TemporaryDirectory(dir=out_dir) as tmpdir:
#         base._download_annotation(
#             genomes_dir=tmpdir, annot_url=annot_url, localname=localname
#         )
#
#         fname = os.path.join(tmpdir, localname, localname + ".annotation.gtf")
#         validate_annot(fname, "gtf")
#
#         fname = os.path.join(tmpdir, localname, localname + ".annotation.bed")
#         validate_annot(fname, "bed")


@pytest.mark.skipif(not travis, reason="slow")
def test_download_annotation(ncbi):
    out_dir = os.getcwd()
    localname = "my_annot"
    name = "ASM14646v1"
    link = (
        "ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/146/465/"
        "GCF_000146465.1_ASM14646v1/GCF_000146465.1_ASM14646v1_genomic.gff.gz"
    )
    with TemporaryDirectory(dir=out_dir) as tmpdir:
        ncbi.download_annotation(name=name, genomes_dir=tmpdir, localname=localname)

        # check download_and_generate_annotation output
        fname = os.path.join(tmpdir, localname, localname + ".annotation.gtf")
        validate_annot(fname, "gtf")

        fname = os.path.join(tmpdir, localname, localname + ".annotation.bed")
        validate_annot(fname, "bed")

        # check attempt_download_and_report_back output
        readme = os.path.join(tmpdir, localname, "README.txt")
        metadata, lines = genomepy.files.read_readme(readme)
        assert metadata["annotation url"].endswith(link)


def test_head_annotation(ncbi, caplog, capsys):
    ncbi.head_annotation("ASM14646v1", n=1)
    captured = capsys.readouterr().out.strip()

    assert "NCBI" in caplog.text
    assert 'gene_name "Eint_010010";' in captured


def test__search_text(ucsc):
    term = "Ailuropoda melanoleuca"
    assert list(ucsc._search_text("not_in_description")) == []
    assert next(ucsc._search_text(term)) == "ailMel1"


def test__search_accession(ncbi):
    assert list(ncbi._search_accession("not_an_id")) == []
    assert next(ncbi._search_accession("GCA_000004335.1")) == "AilMel_1.0"


def test__search_taxonomy(ucsc):
    assert list(ucsc._search_taxonomy("not_an_id")) == []
    assert next(ucsc._search_taxonomy("9646")) == "ailMel1"


def test_search(ucsc):
    for method in ["ailMel1", "9646", "Ailuropoda melanoleuca"]:
        genome = next(ucsc.search(method))
        assert genome[0] == "ailMel1"
        assert isinstance(genome, tuple)



import filecmp
import os
import shutil
import subprocess as sp
from tempfile import NamedTemporaryFile

import pytest
from pyfaidx import Fasta

import genomepy.files


@pytest.fixture
def zipped_genome():
    tmp = NamedTemporaryFile(suffix=".fa.zip", delete=False)
    shutil.copyfile("tests/data/zip_genome.fa.zip", tmp.name)
    yield tmp.name
    # Remove temp file if it's zipped or not
    if os.path.exists(tmp.name):
        os.unlink(tmp.name)
    if os.path.exists(tmp.name[:-4]):
        os.unlink(tmp.name[:-4])


def test_read_readme():
    metadata, lines = genomepy.files.read_readme("tests/data/not_an_existsing_file")
    assert metadata["name"] == "na"
    assert lines == []

    wd = os.getcwd()
    readme = os.path.join(wd, "README.txt")
    with open(readme, "w") as f:
        f.writelines("provider: asd\n")
        f.writelines(
            "this is a regular line\n  \n  \n \nonly one empty line before me\n"
        )

    metadata, lines = genomepy.files.read_readme(readme)
    assert metadata["provider"] == "asd"
    assert lines == ["this is a regular line", "", "only one empty line before me"]

    os.unlink(readme)


def test_write_readme():
    wd = os.getcwd()
    readme = os.path.join(wd, "README.txt")
    metadata, lines = genomepy.files.read_readme(readme)

    metadata["name"] = "my_cool_genome"
    lines = ["", "I wanted to do some regex, but regex is hard"]
    genomepy.files.write_readme(readme, metadata, lines)

    metadata2, lines2 = genomepy.files.read_readme(readme)
    assert metadata == metadata2
    assert lines == lines2

    os.unlink(readme)


def test_update_readme():
    wd = os.getcwd()
    readme = os.path.join(wd, "README.txt")

    # new data added
    updated_metadata = {"key": "value"}
    extra_lines = ["let's add some text to the lines section"]
    genomepy.files.update_readme(readme, updated_metadata, extra_lines)
    metadata, lines = genomepy.files.read_readme(readme)
    assert metadata["key"] == updated_metadata["key"]
    assert extra_lines[0] in lines

    # new data overwrites
    updated_metadata = {"key": "value2"}
    genomepy.files.update_readme(readme, updated_metadata)
    metadata, lines = genomepy.files.read_readme(readme)
    assert metadata["key"] == updated_metadata["key"]

    os.unlink(readme)


def test_extract_archive(zipped_genome):
    assert os.path.exists(zipped_genome)
    fname = genomepy.files.extract_archive(zipped_genome)
    assert fname.endswith(".fa")
    assert os.path.exists(fname)
    assert not os.path.exists(fname + ".zip")


def test_extract_zip(zipped_genome):
    assert os.path.exists(zipped_genome)
    fname = genomepy.files.extract_zip(zipped_genome)
    assert fname.endswith(".fa")
    assert os.path.exists(fname)
    assert not os.path.exists(fname + ".zip")


def test_extract_gzip(fname="tests/data/small_genome.fa.gz"):
    assert os.path.exists(fname)
    fname = genomepy.files.extract_gzip(fname)
    assert fname.endswith(".fa")
    assert os.path.exists(fname)
    assert not os.path.exists(fname + ".gz")


def test_gzip_and_name(fname="tests/data/small_genome.fa"):
    assert os.path.exists(fname)
    fname = genomepy.files.gzip_and_name(fname)
    assert fname.endswith(".gz")
    assert os.path.exists(fname)
    assert not os.path.exists(fname[:-3])

    fname = genomepy.files.extract_gzip(fname)
    assert fname.endswith(".fa")
    assert os.path.exists(fname)
    assert not os.path.exists(fname + ".gz")


def test_bgzip_and_name(fname="tests/data/small_genome.fa"):
    assert os.path.exists(fname)
    fname = genomepy.files.bgzip_and_name(fname)
    assert fname.endswith(".gz")
    assert os.path.exists(fname)
    assert not os.path.exists(fname[:-3])

    with pytest.raises(sp.CalledProcessError):
        genomepy.files.bgzip_and_name("tests/data/nofile.fa")


def test_extract_tarball():
    fname = "tests/data/tar2.fa.tar.gz"
    outname = "tests/data/tar2.fa"
    genomepy.files.extract_tarball(fname, outfile=outname, concat=True)  # noqa

    assert os.path.exists(outname)
    # tar2.fa is a copy of tar1.fa. Check if they are identical after untarring.
    assert filecmp.cmp(outname, "tests/data/tar1.fa")
    os.unlink(outname)


def test__open():
    # read/write regular file
    a1 = "tests/data/data.annotation.gtf"
    with genomepy.files._open(a1, "w") as gtf:
        gtf.write("regular file")

    with open(a1) as gtf:
        lines1 = gtf.readlines()
    assert lines1 == ["regular file"]

    with genomepy.files._open(a1) as gtf:
        lines2 = gtf.readlines()
    assert lines2 == lines1
    genomepy.files.rm_rf(a1)

    # read/write gzipped file
    a2 = "tests/data/data.annotation.gtf.gz"
    with genomepy.files._open(a2, "w") as gtf:
        gtf.write("gzipped file")

    with pytest.raises(UnicodeDecodeError):
        with open(a2) as gtf:
            gtf.read()

    with genomepy.files._open(a2) as gtf:
        lines = gtf.readlines()
    assert lines == ["gzipped file"]
    genomepy.files.rm_rf(a2)


def test_get_file_info(fname="tests/data/small_genome.fa.gz"):
    ext, gz = genomepy.files.get_file_info(fname)
    assert ext == ".fa" and gz

    ext, gz = genomepy.files.get_file_info(fname[:-2] + "fai")
    assert ext == ".fai" and not gz


def test_glob_ext_files(file="tests/data/small_genome.fa"):
    assert file not in genomepy.files.glob_ext_files("tests/data")
    assert file + ".gz" in genomepy.files.glob_ext_files("tests/data")
    assert len(genomepy.files.glob_ext_files("tests/data", "fake_ext")) == 0
    assert len(genomepy.files.glob_ext_files("tests/data/regexp")) == 1


def test_filter_fasta(fname="tests/data/regexp/regexp.fa"):
    # function proper
    regexps = [
        ("Chr.*", 2, 15),
        ("Scaffold.*", 1, 16),
        ("scaffold_.*", 3, 14),
        (r">\d+$", 4, 13),
        ("chr.*", 4, 13),
    ]
    for regex, match, no_match in regexps:
        with NamedTemporaryFile(suffix=".fa") as tmpfa:
            genomepy.files.filter_fasta(
                fname, regex=regex, invert_match=False, outfa=tmpfa.name
            )
            keys = Fasta(tmpfa.name).keys()
            assert len(keys) == match, regex

        with NamedTemporaryFile(suffix=".fa") as tmpfa:
            genomepy.files.filter_fasta(
                fname, regex=regex, invert_match=True, outfa=tmpfa.name
            )
            keys = Fasta(tmpfa.name).keys()
            assert len(keys) == no_match, regex



import os

import pytest
from appdirs import user_config_dir

import genomepy
import genomepy.utils
from tests import linux, travis


def test_head_annotations(caplog, capsys):
    genomepy.functions.head_annotations("ASM14646v1", provider="ncbi", n=1)
    captured = capsys.readouterr().out.strip()

    assert "NCBI" in caplog.text
    assert 'gene_name "Eint_010010";' in captured


def test_list_available_genomes():
    g = genomepy.functions.list_available_genomes("Ensembl")
    metadata = next(g)
    assert isinstance(metadata, list)
    # this will give an error if ensembl adds a new assembly to the top of their list
    assert metadata[0:2] == ["ASM394721v1", "Ensembl"]


def test_list_installed_genomes():
    assert isinstance(genomepy.functions.list_installed_genomes(os.getcwd()), list)

    gdir = os.path.join(os.getcwd(), "tests", "data")
    genomes = genomepy.functions.list_installed_genomes(gdir)
    assert set(genomes) == {
        "regexp",
        "sacCer3",
        "sanitize",
    }  # OSX likes to sort differently

    empty_list = genomepy.functions.list_installed_genomes("./thisdirdoesnotexist")
    assert empty_list == []


def test__lazy_provider_selection():
    # UCB_Xtro_10.0 can be found on both Ensembl and NCBI.
    # Ensembl is first in lazy selection.

    # find genome in specified provider (NCBI)
    name = "UCB_Xtro_10.0"
    provider = "NCBI"
    p = genomepy.functions._lazy_provider_selection(name, provider)
    assert "ncbi" in str(p)

    # GENCODE's FTP does not work on Travis-Linux
    if not (travis and linux):
        # find the first provider (Ensembl)
        provider = None
        p = genomepy.functions._lazy_provider_selection(name, provider)
        assert "ensembl" in str(p)

        # cant find genome anywhere
        name = "not_a_genome"
        with pytest.raises(genomepy.exceptions.GenomeDownloadError):
            genomepy.functions._lazy_provider_selection(name, provider)


def test__provider_selection():
    # specified provider
    name = "UCB_Xtro_10.0"
    localname = "test_genome"
    genomes_dir = os.getcwd()
    provider = "NCBI"
    p = genomepy.functions._provider_selection(name, localname, genomes_dir, provider)
    assert "ncbi" in str(p)

    # lazy provider
    p = genomepy.functions._provider_selection(name, localname, genomes_dir)
    assert "ensembl" in str(p)


def test__provider_selection_readme():
    name = "Xenopus_tropicalis_v9.1"
    localname = "test_genome"
    genomes_dir = os.getcwd()
    provider = "NCBI"
    # provider from readme
    readme = os.path.join(genomes_dir, localname, "README.txt")
    os.makedirs(os.path.dirname(readme), exist_ok=True)
    with open(readme, "w") as r:
        r.write("provider: NCBI")
    provider = None
    p = genomepy.functions._provider_selection(name, localname, genomes_dir, provider)
    assert "ncbi" in str(p)
    genomepy.utils.rm_rf(os.path.dirname(readme))


def test__get_fasta_regex_func():
    # filter alt regions (default)
    func = genomepy.functions._get_fasta_regex_func(regex=None, keep_alt=False)
    assert func("alt1") is False
    assert func("chr1") is True
    assert func("ALT1") is False  # case insensitive

    # filter user specified regex
    func = genomepy.functions._get_fasta_regex_func(
        regex="chr", invert_match=False, keep_alt=True
    )
    assert func("chr1") is True
    assert func("alt1") is False
    assert func("something_else") is False

    # filter user specified regex (inverted)
    func = genomepy.functions._get_fasta_regex_func(
        regex="chr", invert_match=True, keep_alt=True
    )
    assert func("chr1") is False
    assert func("alt1") is True
    assert func("something_else") is True

    # filter both
    func = genomepy.functions._get_fasta_regex_func(
        regex="chr", invert_match=True, keep_alt=False
    )
    assert func("chr1") is False
    assert func("alt1") is False
    assert func("something_else") is True


def test_install_genome():
    localname = "my_genome"
    genomepy.functions.install_genome(
        name="tests/data/sacCer3/sacCer3.fa",
        provider="Local",
        genomes_dir=None,
        localname=localname,
        regex="chrIV",
        annotation=True,
        force=True,
    )

    genomes_dir = genomepy.functions.get_genomes_dir(None, False)
    genome_file = os.path.join(genomes_dir, localname, localname + ".fa")
    assert os.path.exists(genome_file)
    sizes_file = os.path.join(genomes_dir, localname, localname + ".fa.sizes")
    assert os.path.exists(sizes_file)
    gaps_file = os.path.join(genomes_dir, localname, localname + ".gaps.bed")
    assert os.path.exists(gaps_file)
    annotation_file = os.path.join(
        genomes_dir, localname, localname + ".annotation.gtf"
    )
    assert os.path.exists(annotation_file)

    # regex test:
    sizes = genomepy.Genome(localname).sizes.keys()
    assert "chrIV" in sizes


# already used, but we had to install a genome first to test it
def test_generate_exports():
    exports = genomepy.functions._generate_exports()
    assert isinstance(exports, list)
    # check if my_genome was installed in the last test
    assert any([x for x in exports if x.startswith("export MY_GENOME")])

    # add genome that throws an IndexNotFoundError
    gd = genomepy.utils.get_genomes_dir(None, True)
    name = "testgenome"
    os.makedirs(os.path.join(gd, name), exist_ok=True)
    path = os.path.join(gd, name, f"{name}.fa")
    with open(path, "w") as fa:
        fa.write("genome without index")
    exports = genomepy.functions._generate_exports()
    assert f"export TESTGENOME={path}" not in exports

    # add genome that works
    with open(path, "w") as fa:
        fa.write(">chr1\nallowed characters")
    genomepy.Genome(name, gd)  # create index
    exports = genomepy.functions._generate_exports()
    assert f"export TESTGENOME={path}" in exports

    genomepy.utils.rm_rf(os.path.join(gd, "testgenome"))


# already used, but we had to install a genome first to test it
def test_generate_env():
    config_dir = str(user_config_dir("genomepy"))
    path = os.path.join(config_dir, "exports.txt")

    # give file path
    my_path = "~/exports.txt"
    genomepy.functions.generate_env(my_path)
    assert os.path.exists(os.path.expanduser(my_path))
    os.unlink(os.path.expanduser(my_path))

    # give file name
    my_file = os.path.join(config_dir, "my_exports.txt")
    genomepy.functions.generate_env("my_exports.txt")
    assert os.path.exists(my_file)
    os.unlink(os.path.expanduser(my_file))

    # give nothing
    if os.path.exists(path):
        os.unlink(path)
    genomepy.functions.generate_env()
    assert os.path.exists(path)

    with open(path) as f:
        exports = []
        for line in f.readlines():
            vals = line.strip()
            exports.append(vals)
    assert any([x for x in exports if x.startswith("export MY_GENOME")])
    os.unlink(path)


def test__delete_extensions():
    fpath1 = "tests/data/empty/weird_ext1.test123"
    fpath2 = "tests/data/empty/weird_ext2.test123"
    for fpath in [fpath1, fpath2]:
        with open(fpath, "w") as f:
            f.write("asd\n")

    assert os.path.exists(fpath1)
    assert os.path.exists(fpath2)
    genomepy.functions._delete_extensions("tests/data/empty", ["test123"])
    assert not os.path.exists(fpath1)
    assert not os.path.exists(fpath2)


def test__is_genome_dir():
    # dir contains a fasta
    assert genomepy.functions._is_genome_dir("tests/data/regexp")
    # dir does not contain a fasta
    assert not genomepy.functions._is_genome_dir("tests/genome")



"""Genome class, modules & related functions"""
import os.path
import re
from glob import glob

from pyfaidx import Fasta

from genomepy.files import glob_ext_files, read_readme
from genomepy.genome.sequences import get_random_sequences as _get_random_sequences
from genomepy.genome.sequences import track2fasta as _track2fasta
from genomepy.plugins import get_active_plugins
from genomepy.utils import cleanpath, get_genomes_dir, safe

__all__ = ["Genome", "generate_fa_sizes", "generate_gap_bed"]


class Genome(Fasta):
    """
    pyfaidx Fasta object of a genome with additional attributes & methods.

    Generates a genome index file, sizes file and gaps file of the genome.

    Parameters
    ----------
    name : str
        Genome name

    genomes_dir : str, optional
        Genome installation directory

    Returns
    -------
    pyfaidx.Fasta
        An object that provides a pygr compatible interface.
    """

    # import methods
    get_random_sequences = _get_random_sequences
    track2fasta = _track2fasta

    # lazy attributes (loaded when called)
    # listed here for code autocompletion
    sizes: dict = None
    "contents of the sizes file: contigs and their lengths"
    gaps: dict = None
    "contents of the gaps file: contigs and the number of Ns contained"

    def __init__(self, name, genomes_dir=None, *args, **kwargs):
        self.name = safe(os.path.basename(re.sub(r"\.fa(\.gz)?$", "", name)))
        "genome name"
        self.genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)
        "path to the genomepy genomes directory"
        self.filename = self._parse_filename(name)
        super(Genome, self).__init__(self.filename, *args, **kwargs)

        # file paths
        self.genome_file = self.filename
        "path to the genome fasta"
        self.genome_dir = os.path.dirname(self.filename)
        "path to the genome directory"
        self.index_file = self.genome_file + ".fai"
        "path to the genome index"
        self.sizes_file = self._check_support_file("sizes")
        "path to the chromosome sizes file"
        self.gaps_file = self._check_support_file("gaps")
        "path to the chromosome gaps file"
        self.annotation_gtf_file = self._check_annotation_file("gtf")
        "path to the gene annotation GTF file"
        self.annotation_bed_file = self._check_annotation_file("bed")
        "path to the gene annotation BED file"
        self.readme_file = os.path.join(self.genome_dir, "README.txt")
        "path to the README file"

        # genome attributes
        metadata, _ = read_readme(self.readme_file)
        self.tax_id = metadata["tax_id"]
        "genome taxonomy identifier"
        self.assembly_accession = metadata["assembly_accession"]
        "genome assembly accession"

    # lazy attributes
    def __getattribute__(self, name):
        val = super(Genome, self).__getattribute__(name)
        if val is not None:
            return val

        # if the attribute is None/empty, check if it is a lazy attribute
        if name == "sizes":
            val = {}
            with open(self.sizes_file) as f:
                for line in f:
                    contig, length = line.strip().split("\t")
                    val[contig] = int(length)
            setattr(self, name, val)

        elif name == "gaps":
            val = {}
            with open(self.gaps_file) as f:
                for line in f:
                    chrom, start, end = line.strip().split("\t")
                    val[chrom] = val.get(chrom, 0) + int(end) - int(start)
            setattr(self, name, val)

        return val

    # lazily update attributes if upstream attribute is updated
    def __setattr__(self, name, value):
        if name == "sizes_file":
            self.sizes = None  # noqa
        elif name == "gaps_file":
            self.gaps = None  # noqa
        super(Genome, self).__setattr__(name, value)

    @property
    def plugin(self):
        """dict of all active plugins and their properties"""
        p = dict()
        for plugin in get_active_plugins():
            p[plugin.name] = plugin.get_properties(self)
        return p

    def _parse_filename(self, name: str) -> str:
        """
        accepts path to a fasta file, path to a fasta folder, or
        the name of a genome (e.g. hg38).

        returns the abspath to the fasta file
        """
        path_name = cleanpath(name)
        if os.path.isfile(path_name):
            return path_name

        default_genome_dir = os.path.join(self.genomes_dir, self.name)
        for f in glob_ext_files(path_name) + glob_ext_files(default_genome_dir):
            if self.name + ".fa" in os.path.basename(f):
                return f

        raise FileNotFoundError(
            f"could not find {self.name}.fa(.gz) in genome_dir {default_genome_dir}"
        )

    def _check_support_file(self, ftype):
        """generate support file if missing/outdated"""
        ext = ".fa.sizes" if ftype == "sizes" else ".gaps.bed"
        func = generate_fa_sizes if ftype == "sizes" else generate_gap_bed

        fname = os.path.join(self.genome_dir, self.name + ext)
        if not os.path.exists(fname) or os.path.getmtime(fname) < os.path.getmtime(
            self.genome_file
        ):
            func(self.genome_file, fname)
        return fname

    def _check_annotation_file(self, ext):
        """returns (gzipped) annotation file"""
        pattern = os.path.join(self.genome_dir, self.name + ".annotation.")
        file = glob(pattern + ext + "*")
        return file[0] if file else None


def generate_gap_bed(fname, outname):
    """Generate a BED file with gap locations.

    Parameters
    ----------
    fname : str
        Filename of input FASTA file.

    outname : str
        Filename of output BED file.
    """
    f = Fasta(fname)
    with open(outname, "w") as bed:
        for chrom in f.keys():
            for m in re.finditer(r"N+", f[chrom][:].seq):
                bed.write(f"{chrom}\t{m.start(0)}\t{m.end(0)}\n")


def generate_fa_sizes(fname, outname):
    """Generate a fa.sizes file.

    Parameters
    ----------
    fname : str
        Filename of input FASTA file.

    outname : str
        Filename of output BED file.
    """
    f = Fasta(fname)
    with open(outname, "w") as sizes:
        for seqname, seq in f.items():
            sizes.write(f"{seqname}\t{len(seq)}\n")



"""Annotation class, modules & related functions"""
import os
import re
from pathlib import Path
from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd
from loguru import logger

from genomepy.annotation.mygene import _map_genes, query_mygene
from genomepy.annotation.sanitize import _sanitize
from genomepy.annotation.utils import _check_property, _parse_annot, read_annot
from genomepy.files import read_readme
from genomepy.providers import map_locations
from genomepy.utils import cleanpath, get_genomes_dir, safe

__all__ = ["Annotation", "query_mygene", "filter_regex"]


class Annotation:
    """
    Manipulate genes and whole gene annotations with pandas dataframes.

    Parameters
    ----------
    name : str
        Genome name/directory/fasta or gene annotation BED/GTF file.
    genomes_dir : str, optional
        Genomes installation directory.
    quiet : bool, optional
        Silence init warnings

    Returns
    -------
    object
        attributes & methods to manipulate gene annotations
    """

    # import methods
    map_genes = _map_genes
    sanitize = _sanitize

    # lazy attributes (loaded when called)
    # listed here for code autocompletion
    bed: pd.DataFrame = None
    "Dataframe with BED format annotation"
    gtf: pd.DataFrame = None
    "Dataframe with GTF format annotation"
    named_gtf: pd.DataFrame = None
    "Dataframe with GTF format annotation, with gene_name as index"
    genome_contigs: list = None
    "Contigs found in the genome fasta"
    annotation_contigs: list = None
    "Contigs found in the gene annotation BED"

    def __init__(self, name: str, genomes_dir: str = None, quiet: bool = False):
        # name and directory
        n, g = _get_name_and_dir(name, genomes_dir)
        self.name = n
        "genome name"
        self.genome_dir = g
        "path to the genome directory"

        # annotation files
        fname = cleanpath(name)
        suffixes = Path(fname).suffixes[-2:]
        b = fname
        if not (".bed" in suffixes or ".BED" in suffixes):
            b = _get_file(self.genome_dir, f"{self.name}.annotation.bed", not quiet)
        self.annotation_bed_file = b
        "path to the gene annotation BED file"
        g = fname
        if not (".gtf" in suffixes or ".GTF" in suffixes):
            g = _get_file(self.genome_dir, f"{self.name}.annotation.gtf", not quiet)
        self.annotation_gtf_file = g
        "path to the gene annotation GTF file"

        # genome files
        g = fname
        if ".fa" not in suffixes:
            g = _get_file(self.genome_dir, f"{self.name}.fa", False)
        self.genome_file = g
        "path to the genome fasta"
        self.readme_file = _get_file(self.genome_dir, "README.txt", False)
        "path to the README file"
        self.index_file = _get_file(self.genome_dir, f"{self.name}.fa.fai", False)
        "path to the genome index"
        self.sizes_file = _get_file(self.genome_dir, f"{self.name}.fa.sizes", False)
        "path to the chromosome sizes file"

        # genome attributes
        t = read_readme(str(self.readme_file))[0]["tax_id"]
        self.tax_id = None if t == "na" else int(t)
        "genome taxonomy identifier"

    # lazy attributes
    def __getattribute__(self, name):
        val = super(Annotation, self).__getattribute__(name)
        if val is not None:
            return val

        # if the attribute is None/empty, check if it is a lazy attribute
        if name == "bed":
            _check_property(self.annotation_bed_file, f"{self.name}.annotation.bed")
            val = read_annot(self.annotation_bed_file)
            setattr(self, name, val)

        elif name == "gtf":
            _check_property(self.annotation_gtf_file, f"{self.name}.annotation.gtf")
            val = read_annot(self.annotation_gtf_file)
            setattr(self, name, val)

        elif name == "named_gtf":
            val = self.gtf.join(
                self.from_attributes("gene_name"), how="inner"
            ).set_index("gene_name")
            setattr(self, name, val)

        elif name == "genome_contigs":
            _check_property(self.sizes_file, f"{self.name}.fa.sizes")
            val = list(
                set(pd.read_csv(self.sizes_file, sep="\t", header=None, dtype=str)[0])
            )
            setattr(self, name, val)

        elif name == "annotation_contigs":
            val = list(set(self.bed.chrom))
            setattr(self, name, val)

        return val

    # lazily update attributes if upstream attribute is updated
    def __setattr__(self, name, value):
        if name == "bed":
            self.annotation_contigs = None  # noqa
        elif name == "gtf":
            self.named_gtf = None  # noqa
        elif name == "sizes_file":
            self.genome_contigs = None  # noqa
        super(Annotation, self).__setattr__(name, value)

    def from_attributes(
        self, field, annot: Union[str, pd.DataFrame] = "gtf", check=True
    ):
        """
        Convert the specified GTF attribute field to a pandas series

        Parameters
        ----------
        field : str
            field from the GTF's attribute column.
        annot : str or pd.Dataframe, optional
            any GTF in dataframe format, or the default GTF.
        check : bool, optional
            filter the GTF for rows containing field?

        Returns
        -------
        pd.Series
            with the same index as the input GTF and the field column
        """
        df = _parse_annot(self, annot)
        if check:
            df = df[df["attribute"].str.contains(field)]
            if len(df) == 0:
                raise ValueError(f"{field} not in GTF attributes!")

        # extract the text between the quotes
        series = df["attribute"].str.extract(rf'{field} "(.*?)"', expand=False)
        series.name = field
        return series

    def genes(self, annot: str = "gtf") -> list:
        """
        Retrieve gene names from an annotation.

        For BED files, names are taken from the 'name' columns.

        For GTF files, names are taken from the 'gene_name' field
        in the attribute column, if available.

        Parameters
        ----------
        annot : str, optional
            Annotation file type: 'bed' or 'gtf' (default: "gtf")

        Returns
        -------
        list
            gene names
        """
        if annot.lower() == "bed":
            return list(set(self.bed.name))
        return list(set(self.named_gtf.index))

    def gene_coords(self, genes: Iterable[str], annot: str = "bed") -> pd.DataFrame:
        """
        Retrieve gene locations.

        Parameters
        ----------
        genes : Iterable
            List of gene names as found in the given annotation file type
        annot : str, optional
            Annotation file type: 'bed' or 'gtf' (default: "bed")

        Returns
        -------
        pandas.DataFrame
            gene annotation
        """
        gene_list = list(genes)
        if annot.lower() == "bed":
            df = self.bed.set_index("name")
            gene_info = df[["chrom", "start", "end", "strand"]]
        else:
            df = self.named_gtf
            # 1 row per gene
            df = (
                df.groupby(["gene_name", "seqname", "strand"])
                .agg({"start": np.min, "end": np.max})
                .reset_index(level=["seqname", "strand"])
            )
            gene_info = df[["seqname", "start", "end", "strand"]]

        gene_info = gene_info.reindex(gene_list).dropna()
        pct = int(100 * len(set(gene_info.index)) / len(gene_list))
        if pct < 90:
            logger.warning(
                (f"Only {pct}% of genes was found. " if pct else "No genes found. ")
                + "A list of all gene names can be found with `Annotation.genes()`"
            )

        if annot.lower() == "bed":
            return gene_info.reset_index()[["chrom", "start", "end", "name", "strand"]]
        else:
            return gene_info.reset_index()[
                ["seqname", "start", "end", "gene_name", "strand"]
            ]

    def map_locations(
        self, annot: Union[str, pd.DataFrame], to: str, drop=True
    ) -> Union[None, pd.DataFrame]:
        """
        Map chromosome mapping from one assembly to another.

        Uses the NCBI assembly reports to find contigs.
        Drops missing contigs.

        Parameters
        ----------
        annot : str or pd.Dataframe
            annotation to map: "bed", "gtf" or a pandas dataframe.
        to: str
            target provider (UCSC, Ensembl or NCBI)
        drop: bool, optional
            if True, replace the chromosome column.
            If False, add a 2nd chromosome column.

        Returns
        -------
        pandas.DataFrame
            chromosome mapping.
        """
        if self.readme_file is None:
            raise AttributeError(
                "Can only map genomepy annotations (a readme file is required)"
            )
        genomes_dir = os.path.dirname(self.genome_dir)
        mapping = map_locations(self.name, to, genomes_dir)
        if mapping is None:
            return

        df = _parse_annot(self, annot)
        index_name = df.index.name
        if not set([index_name] + df.columns.to_list()) & {"chrom", "seqname"}:
            raise ValueError(
                "Location mapping requires a column named 'chrom' or 'seqname'."
            )

        # join mapping on chromosome column and return with original index
        is_indexed = df.index.to_list() != list(range(df.shape[0]))
        if is_indexed:
            df = df.reset_index(level=index_name)
        index_col = "chrom" if "chrom" in df.columns else "seqname"
        df = df.set_index(index_col)
        df = mapping.join(df, how="inner")
        df = df.reset_index(drop=drop)
        df.columns = [index_col] + df.columns.to_list()[1:]
        if is_indexed:
            df = df.set_index(index_name if index_name else "index")

        return df

    def filter_regex(
        self,
        annot: Union[str, pd.DataFrame],
        regex: Optional[str] = ".*",
        invert_match: Optional[bool] = False,
        column: Union[str, int] = 0,
    ) -> pd.DataFrame:
        """
        Filter a dataframe by any column using regex.

        Parameters
        ----------
        annot : str or pd.Dataframe
            annotation to filter: "bed", "gtf" or a pandas dataframe
        regex : str
            regex string to match
        invert_match : bool, optional
            keep contigs NOT matching the regex string
        column: str or int, optional
            column name or number to filter (default: 1st, contig name)

        Returns
        -------
        pd.DataFrame
            filtered dataframe
        """
        df = _parse_annot(self, annot)
        return filter_regex(df, regex, invert_match, column)

    def gtf_dict(
        self, key, value, string_values=True, annot: Union[str, pd.DataFrame] = "gtf"
    ):
        """
        Create a dictionary based on the columns or attribute fields in a GTF.

        Parameters
        ----------
        key : str
            column name or attribute fields (e.g. "seqname", "gene_name")
        value : str
            column name or attribute fields (e.g. "gene_id", "transcript_name")
        string_values : bool, optional
            attempt to format the dict values as strings
            (only happens if all value lists are length 1)
        annot : str or pd.Dataframe, optional
            annotation to filter: "gtf" or a pandas dataframe

        Returns
        -------
        dict
            with values as lists. If string_values is True
            and all lists are length 1, values will be strings.
        """
        df = _parse_annot(self, annot)
        k = key in df.columns
        v = value in df.columns
        # pd.DataFrame.iterrows() is slow. this is not.
        attributes = zip(
            df[key] if k else df.attribute,
            df[value] if v else df.attribute,
        )

        def _get_attr_item(series, item):
            """
            example series.attribute: "...; gene_name: "TP53"; ..."
            item="gene_name" would return "TP53"
            """
            split = series.split(item)
            return split[1].split('"')[1]  # item might not exist

        def _get_col_item(series, _):
            return series

        get_key = _get_col_item if k else _get_attr_item
        get_val = _get_col_item if v else _get_attr_item
        a_dict = dict()
        for row in attributes:
            try:
                k = get_key(row[0], key)
                v = get_val(row[1], value)
            except IndexError:
                continue

            # unique values per key
            if k in a_dict:
                a_dict[k].update({v})
            else:
                a_dict[k] = {v}

        # return values as str if all values are length 1
        # and string_values is True, else return values as list
        all_len_1 = string_values and all(len(v) == 1 for v in a_dict.values())
        for k, v in a_dict.items():
            a_dict[k] = list(v)[0] if all_len_1 else list(v)

        return a_dict

    def lengths(self, attribute="gene_name"):
        """
        Return a series with the selected GTF attribute as index,
        and its lengths as values.

        Parameters
        ----------
        attribute : str
            attribute to provide lengths of. Options: gene_name, gene_id,
            transcript_name, transcript_id. Attribute must be present in the GTF file.

        Returns
        -------
        pd.Series
            attribute indexed series named 'length'
        """
        exons = self.gtf[self.gtf["feature"] == "exon"].copy()
        if len(exons) == 0:
            raise ValueError("no exon details in GTF!")
        exons = exons.join(self.from_attributes("transcript_id", exons), how="inner")

        # add exon lengths
        exons["length"] = abs(exons["end"] - exons["start"])
        # sum exon lengths per transcript
        lens = exons.groupby("transcript_id")["length"].sum()
        if attribute == "transcript_id":
            return lens

        # add the attribute
        exons = exons.join(self.from_attributes(attribute, exons), how="inner")
        exons = exons[["transcript_id", attribute]].drop_duplicates()
        exons = exons.set_index("transcript_id").join(lens, how="inner")

        # select longest transcript
        lens = exons.groupby(attribute)["length"].max()
        return lens


def _get_name_and_dir(name, genomes_dir=None):
    """
    Returns the name and directory of the genome.
    """
    fname = cleanpath(name)
    genomes_dir = get_genomes_dir(genomes_dir, check_exist=False)
    if os.path.isfile(fname):
        exts = ["gtf", "GTF", "bed", "BED", "fa"]
        if not any(ext in fname for ext in exts):
            raise NotImplementedError(
                "Only (gzipped) bed, gtf or fasta files are supported!"
            )
        genome_dir = os.path.dirname(fname)
        name = safe(os.path.basename(fname))
        # remove suffices
        any_ext = "(" + ")|(".join(exts) + ")"
        name = re.sub(rf"(\.annotation)?\.({any_ext})(\.gz)?$", "", name)
    elif os.path.isdir(fname):
        genome_dir = fname
        name = safe(os.path.basename(fname))
    elif name in os.listdir(genomes_dir):
        genome_dir = os.path.join(genomes_dir, name)
    else:
        raise FileNotFoundError(f"Could not find {name}")
    return name, genome_dir


def _get_file(genome_dir: str, fname: str, warn_missing: Optional[bool] = True):
    """
    Returns the filepath to a single (gzipped) file in the genome_dir with matching ext.
    """
    fpath = os.path.join(genome_dir, fname)
    if os.path.exists(fpath):
        return fpath
    if os.path.exists(f"{fpath}.gz"):
        return f"{fpath}.gz"
    if warn_missing:
        logger.warning(
            f"Could not find '{fname}(.gz)' in directory {genome_dir}. "
            "Methods using this file won't work!"
        )
    return


def filter_regex(
    df: pd.DataFrame,
    regex: str,
    invert_match: Optional[bool] = False,
    column: Union[str, int] = 0,
) -> pd.DataFrame:
    """
    Filter a pandas dataframe by a column (default: 1st, contig name).

    Parameters
    ----------
    df: pd.Dataframe
        annotation to filter (a pandas dataframe)
    regex : str
        regex string to match
    invert_match : bool, optional
        keep contigs NOT matching the regex string
    column: str or int, optional
        column name or number to filter (default: 1st, contig name)

    Returns
    -------
    pd.DataFrame
        filtered dataframe
    """
    if column not in df.columns:
        if isinstance(column, int):
            column = df.columns[column]
        else:
            raise ValueError(
                f"Column '{column}' not found in annotation columns {list(df.columns)}"
            )

    pattern = re.compile(regex)
    filter_func = df[column].map(lambda x: bool(pattern.match(x)) is not invert_match)
    return df[filter_func]



"""Utility functions with gene annotations"""
import csv
import os
import shutil
import subprocess as sp
from tempfile import mkdtemp

import numpy as np
import pandas as pd

from genomepy.files import _open, extracted_file, gzip_and_name
from genomepy.utils import check_ucsc_tools, rm_rf

GTF_FORMAT = [
    "seqname",
    "source",
    "feature",
    "start",
    "end",
    "score",
    "strand",
    "frame",
    "attribute",
]

BED12_FORMAT = [
    "chrom",
    "start",
    "end",
    "name",
    "score",
    "strand",
    "thickStart",
    "thickEnd",
    "itemRrgb",
    "blockCount",
    "blockSizes",
    "blockStarts",
]


def _check_property(prop, fname):
    if prop is None:
        raise FileNotFoundError(f"'{fname}' required.")


def count_columns(fpath):
    with _open(fpath) as f:
        for line in f:
            if not line.startswith("#"):
                columns = len(line.split("\t"))
                return columns


def read_annot(fpath: str) -> pd.DataFrame:
    """
    Read a GTF or BED file to a dataframe.

    Parameters
    ----------
    fpath : str
        path to the annotation file

    Returns
    -------
    pd.DataFrame
        annotation dataframe
    """
    # determine file type by column count
    columns = count_columns(fpath)
    if columns == 9:
        names = GTF_FORMAT
    elif columns == 12:
        names = BED12_FORMAT
    else:
        raise ValueError(f"{columns} columns detected. BED=12, GTF=9.")

    df = pd.read_csv(
        fpath,
        sep="\t",
        names=names,
        dtype={
            "seqname": str,
            "source": str,
            "feature": str,
            "start": np.uint32,
            "end": np.uint32,
            "score": str,  # int or .
            "strand": str,  # +, - or .
            "frame": str,  # int or .
            "attribute": str,
            "chrom": str,
            "name": str,
            # "thickStart",
            # "thickEnd",
            # "itemRrgb",
            # "blockCount",
            # "blockSizes",
            # "blockStarts",
        },
        comment="#",
    )
    return df


def write_annot(df: pd.DataFrame, fpath: str):
    """
    Write a dataframe to a GTF or BED file.

    Parameters
    ----------
    df : pd.DataFrame
        annotation dataframe
    fpath : str
        path to the annotation file
    """
    df.to_csv(
        fpath,
        sep="\t",
        header=False,
        index=False,
        quoting=csv.QUOTE_NONE,
    )


def generate_annot(template, target, overwrite=False):
    """
    Create an annotation file type from the other file type.

    Parameters
    ----------
    template: str
        a GTF or BED filepath.
    target: str
        filepath to save the new annotation to.
    overwrite: bool, optional
        overwrite existing target file?
    """
    check_ucsc_tools()

    exts = os.path.basename(template.lower()).split(".")
    exts = [e for e in exts if e in ["gtf", "bed"]]
    if len(exts) == 0:
        raise ValueError("Template file must be in GTF or BED format.")
    template_ext = exts[-1]

    if not overwrite and os.path.exists(target):
        raise FileExistsError(f"{target} already exists! Set overwrite=True to ignore.")

    target_dir = os.path.dirname(target)
    tmp_dir = mkdtemp(dir=target_dir)
    tmp_target = os.path.join(tmp_dir, "new_annot")

    if template_ext == "bed":
        cmd = "bedToGenePred {0} /dev/stdout | genePredToGtf -source=genomepy file /dev/stdin {1}"
    else:
        cmd = "gtfToGenePred -genePredExt -ignoreGroupsWithoutExons {0} /dev/stdout | genePredToBed /dev/stdin {1}"

    # unzip template if needed
    with extracted_file(template) as _template:
        sp.check_call(cmd.format(_template, tmp_target), shell=True)
        # gzip if needed
        tmp_target = gzip_and_name(tmp_target, target.endswith(".gz"))

    shutil.move(tmp_target, target)
    rm_rf(tmp_dir)


def _parse_annot(self, annot):
    if isinstance(annot, pd.DataFrame):
        df = annot
    elif isinstance(annot, str) and annot == "bed":
        df = self.bed
    elif isinstance(annot, str) and annot == "gtf":
        df = self.gtf
    else:
        raise ValueError("Argument 'annot' must be 'gtf', 'bed' or a pandas dataframe.")
    return df



import os
from shutil import copyfile
from tempfile import TemporaryDirectory

import pytest

import genomepy.exceptions


def test_ucscprovider(ucsc):
    assert ucsc.name == "UCSC"
    assert ucsc.taxid_fields == ["taxId"]


def test__search_accession(ucsc):
    assert list(ucsc._search_accession("not_an_id")) == []
    assert next(ucsc._search_accession("GCA_000004335.1")) == "ailMel1"


def test__search_accession_ncbi(ucsc):
    ret = list(ucsc._search_accession_ncbi("GCF_000004195.2"))
    expected = ["xenTro1", "xenTro10", "xenTro2", "xenTro3", "xenTro7", "xenTro9"]
    assert ret == expected


def test_assembly_accession(ucsc):
    # in UCSC(, not in NCBI)
    accession = ucsc.assembly_accession("sacCer3")
    assert accession == "GCA_000146045.2"

    # not in UCSC, not in NCBI
    accession = ucsc.assembly_accession("sacCer1")
    assert accession is None

    # not in UCSC, in NCBI
    accession = ucsc.assembly_accession("xenTro7")
    assert accession == "GCA_000004195.2"


def test_annotation_links(ucsc):
    annots = ucsc.annotation_links("sacCer3")
    expected = ["ensGene", "ncbiRefSeq"]
    assert sorted(annots) == sorted(expected)


def test_genome_info_tuple(ucsc):
    t = ucsc._genome_info_tuple("sacCer3")
    assert isinstance(t, tuple)
    assert t[0:4] == ("sacCer3", "GCA_000146045.2", 559292, [True, False, True, False])


def test_get_genome_download_link(ucsc):
    link = ucsc.get_genome_download_link("sacCer3", mask="soft")
    assert link in [
        "http://hgdownload.soe.ucsc.edu/goldenPath/sacCer3/bigZips/chromFa.tar.gz",
        "http://hgdownload.soe.ucsc.edu/goldenPath/sacCer3/bigZips/sacCer3.fa.gz",
    ]

    link = ucsc.get_genome_download_link("danRer7", mask="hard")
    assert link in [
        "http://hgdownload.soe.ucsc.edu/goldenPath/danRer7/bigZips/chromFaMasked.tar.gz",
        "http://hgdownload.soe.ucsc.edu/goldenPath/danRer7/bigZips/danRer7.fa.masked.gz",
    ]


def test__post_process_download(ucsc):
    localname = "tmp"
    out_dir = os.getcwd()
    with TemporaryDirectory(dir=out_dir) as tmpdir:
        # this should skip without error
        ucsc._post_process_download(name=None, fname="", out_dir=None, mask="soft")
        ucsc._post_process_download(name=None, fname="", out_dir=None, mask="hard")
        ucsc._post_process_download(name=None, fname="", out_dir=None, mask="???")

        # copy fa file for unmasking
        g = os.path.join(tmpdir, localname + ".fa")
        copyfile("tests/data/gap.fa", g)

        ucsc._post_process_download(name=None, fname=g, out_dir=None, mask="none")
        assert os.path.exists(g)
        with open(g) as f:
            for line in f:
                assert "a" not in line


def test_get_annotation_download_links(ucsc):
    # any GTF format annotation
    genome = "sacCer3"
    annots = ucsc.get_annotation_download_links(genome)
    annots.sort()
    expected = ["ensGene", "ncbiRefSeq"]
    assert annots == expected

    # no annotation available
    assert ucsc.get_annotation_download_links("apiMel1") == []


def test_get_annotation_download_link(ucsc):
    # any annotation
    genome = "xenTro2"
    annot = ucsc.get_annotation_download_link(genome)
    assert annot == "refGene"

    # specific annotation type
    genome = "sacCer3"
    annot = ucsc.get_annotation_download_link(
        genome, **{"ucsc_annotation_type": "ncbiRefSeq"}
    )
    assert annot == "ncbiRefSeq"

    # no annotation available
    with pytest.raises(genomepy.exceptions.GenomeDownloadError):
        ucsc.get_annotation_download_link("apiMel1")

    # annotation type non-existing for this genome
    with pytest.raises(FileNotFoundError):
        ucsc.get_annotation_download_link(genome, **{"ucsc_annotation_type": "refGene"})

    # annotation type non-existing
    with pytest.raises(FileNotFoundError):
        ucsc.get_annotation_download_link(genome, **{"ucsc_annotation_type": "what?"})


def test_download_annotation(ucsc):
    ucsc.download_annotation("sacCer1")


def test_get_genomes(ucsc):
    assert isinstance(ucsc.genomes, dict)
    assert "ailMel1" in ucsc.genomes
    genome = ucsc.genomes["ailMel1"]
    assert isinstance(genome, dict)
    for field in ucsc.accession_fields + ucsc.taxid_fields + ucsc.description_fields:
        assert field in genome
    assert genome["taxId"] == 9646
    assert genome["assembly_accession"] == "GCF_000004335.2"
    assert sorted(genome["annotations"]) == sorted(["ncbiRefSeq", "ensGene"])


def test_head_annotation(ucsc, caplog, capsys):
    ucsc.head_annotation("hg38", n=1, **{"annotations": ["ncbiRefSeq", "refGene"]})
    captured = capsys.readouterr().out.strip()

    assert "ncbiRefSeq" in caplog.text
    assert 'gene_name "DDX11L1";' in captured

    assert "refGene" in caplog.text
    assert 'gene_name "WASH7P";' in captured



import argparse
from tempfile import TemporaryDirectory

import pytest

import genomepy.argparse_support


def test_argparse_plugin():
    action = genomepy.argparse_support.parse_genome

    with TemporaryDirectory() as tmpdir:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-g", dest="genome", action=action(auto_install=True, genomes_dir=tmpdir)
        )
        args = parser.parse_args(
            ["-g", "ASM2732v1"],
        )
        assert isinstance(args.genome, genomepy.Genome)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        parser = argparse.ArgumentParser()
        parser.add_argument("-g", dest="genome", action=action())
        _ = parser.parse_args(["-g", "non_existing"])
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1



import pandas as pd
import pytest

import genomepy.files
import genomepy.utils
from tests import linux, travis


def test_create():
    u = genomepy.providers.create("url")
    assert u.name == "URL"
    assert hasattr(u, "download_assembly_report")

    with pytest.raises(ValueError):
        genomepy.providers.create("error")


def test_list_providers():
    assert len(genomepy.providers.list_providers()) == len(genomepy.providers.PROVIDERS)


def test_list_online_providers():
    _all = genomepy.providers.list_providers()
    online = genomepy.providers.list_online_providers()
    if linux and travis:
        assert online == _all[1:]  # GENCODE is on FTP
    else:
        assert online == _all


def test_online_providers():
    # provider given
    providers = list(genomepy.providers.online_providers("url"))
    assert len(providers) == 1
    assert providers[0].name == "URL"

    # not given
    providers = list(genomepy.providers.online_providers())
    if linux and travis:
        assert len(providers) == 5  # GENCODE is on FTP
    else:
        assert len(providers) == 6


def test_search():
    # unrecognized provider/genome will cause an exception or stopiteration respectively
    # case insensitive description search
    search = list(genomepy.providers.search("xEnOpUs TrOpIcAlIs", "ensembl"))
    metadata = search[0]

    # case insensitive assembly name search
    search = list(genomepy.providers.search("UcB_xTrO_10.0", "ensembl"))
    metadata2 = search[0]

    assert metadata == metadata2
    assert isinstance(metadata, list)
    assert "UCB_Xtro_10.0" in str(metadata[0])
    assert "Ensembl" in str(metadata[1])
    assert "GCA_000004195" in str(metadata[2])
    assert "8364" in str(metadata[3])


def test_map_location():
    mapping = genomepy.providers.map_locations(
        frm="sacCer3", to="ensembl", genomes_dir="tests/data"
    )
    assert isinstance(mapping, pd.DataFrame)
    assert mapping.index.name == "ucsc_name"
    assert mapping.columns.to_list() == ["ensembl_name"]


def test_provider(provider):
    assert hasattr(provider, "list")
    assert hasattr(provider, "online")
    assert hasattr(provider, "create")



import pytest

import genomepy.providers.url
import genomepy.utils


def test_urlprovider(url):
    assert url.name == "URL"
    assert url.genomes == {}


def test_genome_taxid(url):
    assert url.genome_taxid(None) is None


def test_assembly_accession(url):
    assert url.assembly_accession(None) is None


def test_search(url):
    result = url.search("???")
    with pytest.raises(StopIteration):
        assert next(result)


def test__genome_info_tuple(url):
    assert url._genome_info_tuple(None) is tuple()


def test__check_name(url):
    assert url._check_name(None) is None


def test_get_genome_download_link(url):
    link = url.get_genome_download_link("url")
    assert link == "url"


def test_get_annotation_download_link(url):
    target = "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_Xenbase.gtf"
    link = url.get_annotation_download_link(None, **{"to_annotation": target})
    assert link == target

    with pytest.raises(TypeError):
        bad_url = "bad_url"
        url.get_annotation_download_link(None, **{"to_annotation": bad_url})

    with pytest.raises(TypeError):
        bad_url = "http://good_url.bad_ext"
        url.get_annotation_download_link(None, **{"to_annotation": bad_url})


def test_fuzzy_annotation_search():
    search_name = "mygenome"
    search_list = [
        "nothing.gtf",
        "mygenome.genes.gff3",
        "asdasdasdmygenomeasdasdasd.gtf",
        "mygenome.zipped.gtf.gz",
    ]
    expected = [
        "mygenomeasdasdasd.gtf",
        "mygenome.zipped.gtf.gz",
        "mygenome.genes.gff3",
    ]
    # f"{search_name}.*?\.{ext}3?(\.gz)?"
    hits = genomepy.providers.url.fuzzy_annotation_search(search_name, search_list)
    assert hits == expected


def test_search_url_for_annotations():
    target = "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_genome.fa.gz"
    links = genomepy.providers.url.search_url_for_annotations(target)
    expected = [
        "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_Xenbase.gtf",
        "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_GCA.gff3.gz",
        "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_GCF.gff3.gz",
        "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_Xenbase.gff3.gz",
        "http://ftp.xenbase.org/pub/Genomics/JGI/Xentr9.1/XENTR_9.1_Xenbase_longest.gff3.gz",
    ]
    assert links == expected

    target = (
        "https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/027/325/"
        "GCF_000027325.1_ASM2732v1/GCF_000027325.1_ASM2732v1_genomic.fna.gz"
    )
    links = genomepy.providers.url.search_url_for_annotations(target)
    expected = [
        "https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/027/325/"
        + "GCF_000027325.1_ASM2732v1/GCF_000027325.1_ASM2732v1_genomic.gtf.gz",
        "https://ftp.ncbi.nlm.nih.gov/genomes/all/GCF/000/027/325/"
        + "GCF_000027325.1_ASM2732v1/GCF_000027325.1_ASM2732v1_genomic.gff.gz",
    ]
    assert links == expected

    # no annot file
    target = (
        "http://ftp.ensembl.org/pub/release-100/fasta/marmota_marmota_marmota/"
        "dna/Marmota_marmota_marmota.marMar2.1.dna.toplevel.fa.gz"
    )
    links = genomepy.providers.url.search_url_for_annotations(target)
    assert links == []



import os

from loguru import logger

from genomepy.files import extracted_file
from genomepy.plugins import Plugin
from genomepy.utils import cmd_ok, mkdir_p, rm_rf, run_index_cmd


class StarPlugin(Plugin):
    def after_genome_download(self, genome, threads=1, force=False):
        index_name = genome.plugin["star"]["index_name"]
        if not cmd_ok("STAR") or (os.path.exists(index_name) and not force):
            return

        index_dir = genome.plugin["star"]["index_dir"]
        rm_rf(index_dir)
        mkdir_p(index_dir)

        # gunzip genome if bgzipped and return up-to-date genome name
        with extracted_file(genome.filename) as fname:
            # index command
            cmd = (
                f"STAR --runMode genomeGenerate --runThreadN {threads} "
                + f"--genomeFastaFiles {fname} --genomeDir {index_dir} "
                + f"--outFileNamePrefix {index_dir}"
            )

            # if an annotation is present, generate a splice-aware index
            gtf_file = genome.annotation_gtf_file
            if gtf_file:
                with extracted_file(gtf_file) as _gtf_file:
                    # update index command with annotation
                    cmd += f" --sjdbGTFfile {_gtf_file}"

                    # Create index
                    run_index_cmd("star", cmd)
            else:
                logger.info("Creating STAR index without annotation file.")
                # Create index
                run_index_cmd("star", cmd)

    def get_properties(self, genome):
        props = {
            "index_dir": os.path.join(
                os.path.dirname(genome.filename), "index", "star"
            ),
            "index_name": os.path.join(
                os.path.dirname(genome.filename), "index", "star", "SA"
            ),
        }
        return props



"""Plugin class, modules & related functions"""
import os
import re

from genomepy.config import config

__all__ = ["Plugin", "manage_plugins", "get_active_plugins"]


class Plugin:
    """Plugin base class."""

    def __init__(self):
        self.name = convert(type(self).__name__).replace("_plugin", "")
        self.active = False

    def activate(self):
        self.active = True

    def deactivate(self):
        self.active = False

    def after_genome_download(self, genome, threads, force):
        raise NotImplementedError("plugin should implement this method")

    def get_properties(self, genome):
        raise NotImplementedError("plugin should implement this method")


def convert(name: str) -> str:
    """
    Convert CamelCase to underscore. e.g. StarPlugin -> star_plugin

    Parameters
    ----------
    name : str
        Camelcase string

    Returns
    -------
    name : str
        Converted name
    """
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def list_plugins() -> list:
    plugin_dir = os.path.dirname(os.path.realpath(__file__))
    plugin_files = [f for f in os.listdir(plugin_dir) if f.endswith(".py")]
    plugin_names = [f[:-3] for f in plugin_files if not f.startswith("_")]
    return plugin_names


def init_plugins():
    """
    create a dictionary of plugin instances

    Returns
    -------
    plugins : dictionary
        key is plugin name, value Plugin object
    """
    # import plugins
    for plugin in list_plugins():
        __import__(f"genomepy.plugins.{plugin}")

    # for each Plugin subclass, save an instance to a dict
    d = {}
    active_plugins = config.get("plugin", [])
    for c in Plugin.__subclasses__():
        ins = c()

        if ins.name in active_plugins:
            ins.activate()

        d[ins.name] = ins

    return d


PLUGINS = init_plugins()


def get_active_plugins() -> list:
    """Returns all active plugin instances."""
    return [inst for name, inst in PLUGINS.items() if inst.active]


def activate(name):
    """Activate plugin.

    Parameters
    ----------
    name : str
        Plugin name.
    """
    if name in PLUGINS:
        PLUGINS[name].activate()
    else:
        raise ValueError(f"plugin {name} not found")


def deactivate(name):
    """Deactivate plugin.

    Parameters
    ----------
    name : str
        Plugin name.
    """
    if name in PLUGINS:
        PLUGINS[name].deactivate()
    else:
        raise ValueError(f"plugin {name} not found")


def show_plugins():
    active_plugins = config.get("plugin", [])
    print("{:20}{}".format("plugin", "enabled"))
    for plugin in sorted(PLUGINS):
        print(
            "{:20}{}".format(plugin, {False: "", True: "*"}[plugin in active_plugins])
        )


def manage_plugins(command: str, plugin_names: list = None):
    """
    Manage genomepy plugins

    Parameters
    ----------
    command : str
        command to perform. Options:

        list
            show plugins and status
        enable
            enable plugins
        disable
            disable plugins
    plugin_names : list
        plugin names for the enable/disable command
    """
    if command in ["show", "list"]:
        return show_plugins()

    active_plugins = config.get("plugin", [])
    for name in plugin_names if plugin_names else []:
        if name not in PLUGINS:
            raise ValueError(f"Unknown plugin: '{name}'.")

    if command in ["enable", "activate"]:
        [active_plugins.append(name) for name in plugin_names]

    elif command in ["disable", "deactivate"]:
        [active_plugins.remove(name) for name in plugin_names if name in active_plugins]

    else:
        raise ValueError(
            f"Invalid plugin command: '{command}'. Options: 'list', 'enable' or 'disable'."
        )

    active_plugins = sorted(set(active_plugins))
    config["plugin"] = active_plugins
    config.save()
    print(f"Enabled plugins: {', '.join(active_plugins)}")
